@@ -30,7 +30,7 @@ type int_data = {
   flags : flags;
   program_mode : bool;
   intenv : Constrintern.internalization_env;
-  notations : Vernacexpr.notation_declaration list
+  notations : Vernacexpr.decl_notation list
 }
 
 exception Conflict
@@ -56,7 +56,7 @@ let decompose_rel evd c =
   | _ -> None
 
 let rec unify env evd flex g x y =
-  if is_conv_leq env evd x y then id_subst g
+  if check_conv env evd x y then id_subst g
   else
     match decompose_rel evd x with
     | Some i -> 
@@ -72,8 +72,8 @@ let rec unify env evd flex g x y =
           single_subst env evd i (PInac x) g
         else raise Stuck
       | None ->
-        let (c, l) = decompose_app_list evd x
-        and (c', l') = decompose_app_list evd y in
+        let (c, l) = decompose_app evd x 
+        and (c', l') = decompose_app evd y in
         if isConstruct evd c && isConstruct evd c' then
           if eq_constr evd c c' then
             unify_constrs env evd flex g l l'
@@ -94,17 +94,16 @@ and unify_constrs env evd flex g l l' =
   | [], [] -> id_subst g
   | hd :: tl, hd' :: tl' ->
     (try
-       let hdunif = unify env evd flex g hd hd' in
-       let { src_ctx = d; map_inst = s } = hdunif in
+       let (d,s,_ as hdunif) = unify env evd flex g hd hd' in
        let specrest = List.map (specialize_constr evd s) in
        let tl = specrest tl and tl' = specrest tl' in
        let tlunif = unify_constrs env evd flex d tl tl' in
        compose_subst env ~sigma:evd tlunif hdunif
      with Stuck ->
        let tlunif = unify_constrs env evd flex g tl tl' in
-       let spec = specialize_constr evd (tlunif.map_inst) in
+       let spec = specialize_constr evd (pi2 tlunif) in
        let hd = spec hd and hd' = spec hd' in
-       let hdunif = unify env evd flex (tlunif.src_ctx) hd hd' in
+       let (d, s, _) as hdunif = unify env evd flex (pi1 tlunif) hd hd' in
        compose_subst env ~sigma:evd hdunif tlunif)
   | _, _ -> raise Conflict
 
@@ -171,9 +170,9 @@ and match_patterns env sigma pl l =
 
 open Constrintern
 
-let matches env sigma (p : user_pats) subst = 
+let matches env sigma (p : user_pats) ((phi,p',g') : context_map) = 
   try
-    let p' = filter (fun x -> not (hidden x)) (rev subst.map_inst) in
+    let p' = filter (fun x -> not (hidden x)) (rev p') in
     UnifSuccess (match_patterns env sigma p p')
   with Conflict -> UnifFailure | Stuck -> UnifStuck
 
@@ -206,8 +205,8 @@ and match_user_patterns env pl l =
      | _, _ -> raise Stuck)
   | _ -> raise Conflict
 
-let matches_user env subst (p : user_pats) = 
-  try UnifSuccess (match_user_patterns env (filter (fun x -> not (hidden x)) (rev subst.map_inst)) p)
+let matches_user env ((phi,p',g') : context_map) (p : user_pats) = 
+  try UnifSuccess (match_user_patterns env (filter (fun x -> not (hidden x)) (rev p')) p)
   with Conflict -> UnifFailure | Stuck -> UnifStuck
 
 let refine_arg idx ctx =
@@ -221,7 +220,7 @@ let adjust_sign_arity env evars p clauses =
     match clauses with
     | [] -> Context.Rel.nhyps p.program_sign
     | _ ->
-      List.fold_left (fun acc (Pre_clause (_, lhs, rhs)) ->
+      List.fold_left (fun acc (_, lhs, rhs) ->
           let len = List.length lhs in
           max acc len) 0 clauses
   in
@@ -239,7 +238,7 @@ let adjust_sign_arity env evars p clauses =
           user_err_loc (None, str "Too many patterns in clauses for this type")
     in aux evars max_args [] fullty
   in
-  let check_clause (Pre_clause (loc, lhs, rhs)) =
+  let check_clause (loc, lhs, rhs) =
     if List.length lhs < max_args then user_err_loc (loc, str "This clause has not enough arguments")
     else ()
   in List.iter check_clause clauses;
@@ -256,16 +255,16 @@ let lets_of_ctx env ctx evars s =
         | PRel i -> (ctx', cs, (i, id) :: varsubst, k, Id.Set.add id ids)
         | _ -> 
           let ty = e_type_of envctx evars c in
-          (make_def (nameR id) (Some (lift k c)) (lift k ty) :: ctx', (c :: cs),
+          (make_def (Context.nameR id) (Some (lift k c)) (lift k ty) :: ctx', (c :: cs),
            varsubst, succ k, Id.Set.add id ids))
       ([],[],[],0,Id.Set.empty) s
   in
   let _, _, ctx' = List.fold_right (fun decl (ids, i, ctx') ->
       let (n, b, t) = to_tuple decl in
-      try ids, pred i, (make_def (nameR (List.assoc i varsubst)) b t :: ctx')
+      try ids, pred i, (make_def (Context.nameR (List.assoc i varsubst)) b t :: ctx')
       with Not_found -> 
         let id' = Namegen.next_name_away n.Context.binder_name ids in
-        Id.Set.add id' ids, pred i, (make_def (nameR id') b t :: ctx')) ctx (ids, List.length ctx, [])
+        Id.Set.add id' ids, pred i, (make_def (Context.nameR id') b t :: ctx')) ctx (ids, List.length ctx, [])
   in pats, ctxs, ctx'
 
 let env_of_rhs evars ctx env s lets = 
@@ -307,11 +306,7 @@ let env_of_rhs evars ctx env s lets =
  *   (List.rev ctx', id_pats ctx, ctx) *)
 
 let is_wf_ref id rec_type =
-  let aux = function 
-    | Some (Logical (nargs, (_, id'))) -> 
-      if Id.equal id id' then Some nargs else None 
-    | _ -> None
-  in CList.find_map_exn aux rec_type
+  List.exists (function Some (Logical (_, id')) -> Id.equal id id' | _ -> false) rec_type
 
 let add_wfrec_implicits rec_type c =
   let open Glob_term in
@@ -324,19 +319,15 @@ let add_wfrec_implicits rec_type c =
           | GApp (fn, args) ->
             DAst.with_loc_val (fun ?loc gfn ->
                 match gfn with
-                | GVar fid -> 
-                  (match is_wf_ref fid rec_type with
-                  | exception Not_found -> maprec c
-                  | nargs -> 
-                    let kind = Evar_kinds.{ qm_obligation = Define false;
-                                            qm_name = Anonymous;
-                                            qm_record_field = None }
-                    in
-                    let newarg = GHole (GQuestionMark kind) in
-                    let newarg = DAst.make ?loc newarg in
-                    let before, after = List.chop nargs (mapargs args) in
-                    let args' = List.append before (newarg :: after) in
-                    DAst.make ?loc (GApp (fn, args')))
+                | GVar fid when is_wf_ref fid rec_type ->
+                  let kind = Evar_kinds.{ qm_obligation = Define false;
+                                          qm_name = Anonymous;
+                                          qm_record_field = None }
+                  in
+                  let newarg = GHole (Evar_kinds.QuestionMark kind, Namegen.IntroAnonymous, None) in
+                  let newarg = DAst.make ?loc newarg in
+                  let args' = List.append (mapargs args) [newarg] in
+                  DAst.make ?loc (GApp (fn, args'))
                 | _ -> maprec c) fn
           | _ -> maprec c) c
     in aux c
@@ -376,7 +367,7 @@ let interp_program_body env sigma ctx data body ty =
           | Name na -> (na, mkRel i) :: subst
           | _ -> subst) 1 [] ctx
     in
-    let c = Vars.replace_vars sigma subst c in
+    let c = Vars.replace_vars subst c in
     let sigma =
       match ty with
       | None -> fst (Typing.type_of env sigma c)
@@ -445,22 +436,23 @@ let unify_type env evars before id ty after =
     let ind, params = dest_ind_family indf in
     let vs = List.map (Tacred.whd_simpl envb !evars) args in
     let params = List.map (Tacred.whd_simpl envb !evars) params in
-    let cstrs = Inductiveops.type_of_constructors envb ind in
+    let cstrs = Inductiveops.type_of_constructors envb (from_peuniverses !evars ind) in
     let cstrs = 
       Array.mapi (fun i ty ->
+          let ty = of_constr ty in
           let ty = prod_applist !evars ty params in
-          let ctx, ty = decompose_prod_decls !evars ty in
+          let ctx, ty = decompose_prod_assum !evars ty in
           let ctx = 
             fold_right (fun decl acc ->
-                let open Context.Rel.Declaration in
-                let id = match get_name decl with
-                | Name id -> next_ident_away id
+                let (n, b, t) = to_tuple decl in
+                match n.binder_name with
+                | Name id -> let id' = next_ident_away id in
+                  (make_def (nameR id') b t :: acc)
                 | Anonymous ->
                   let x = Namegen.id_of_name_using_hdchar
-                      (push_rel_context acc envb) !evars (get_type decl) Anonymous in
-                  next_ident_away x
-                in
-                (set_name (Name id) decl :: acc))
+                      (push_rel_context acc envb) !evars t Anonymous in
+                  let id = next_ident_away x in
+                  (make_def (nameR id) b t :: acc))
               ctx []
           in
           let env' = push_rel_context ctx env in
@@ -513,7 +505,7 @@ let unify_type env evars before id ty after =
     equations_debug Pp.(fun () -> str"Unification raised Not_found");
     None
 
-let blockers env curpats (subst : context_map) =
+let blockers env curpats ((_, patcs, _) : context_map) =
   let rec pattern_blockers p c =
     match DAst.get p, c with
     | PUVar (i, _), t -> []
@@ -532,7 +524,7 @@ let blockers env curpats (subst : context_map) =
       (pattern_blockers hd hd') @ (patterns_blockers tl tl')
     | _ -> []
 
-  in patterns_blockers curpats (rev subst.map_inst)
+  in patterns_blockers curpats (rev patcs)
 
 let subst_matches_constr sigma k s c =
   let rec aux depth c =
@@ -561,8 +553,7 @@ let split_var (env,evars) var delta =
   let branch = function
     | UnifFailure -> None
     | UnifStuck -> assert false
-    | UnifSuccess (map, ctxlen, cstr, cstrpat) ->
-      let { src_ctx = ctx'; map_inst = s; tgt_ctx = ctx } = map in
+    | UnifSuccess ((ctx',s,ctx), ctxlen, cstr, cstrpat) ->
       (* ctx' |- s : before ; ctxc *)
       (* ctx' |- cpat : ty *)
       if !debug then Feedback.msg_debug Pp.(str"cpat: " ++ pr_pat env !evars cstrpat);
@@ -571,8 +562,7 @@ let split_var (env,evars) var delta =
       (* ctx' |- spat : before ; id *)
       let spat =
         let ctxcsubst, beforesubst = List.chop ctxlen s in
-        let map = { src_ctx = ctx'; map_inst = cpat :: beforesubst; tgt_ctx = decl :: before } in
-        check_ctx_map env !evars map
+        check_ctx_map env !evars (ctx', cpat :: beforesubst, decl :: before)
       in
       (* ctx' ; after |- safter : before ; id ; after = delta *)
       Some (lift_subst env !evars spat after)
@@ -641,7 +631,7 @@ let instance_of_pats env evars (ctx : rel_context) (pats : (int * bool) list) =
   let pats' =
     List.map_i (fun i id ->
         let i', _ = lookup_named_i id ctx' in
-        CList.find_map_exn (fun (i'', hide) ->
+        CList.find_map (fun (i'', hide) ->
             if i'' == i then Some (if hide then PHide i' else PRel i')
             else None) pats)
       1 subst
@@ -650,11 +640,11 @@ let instance_of_pats env evars (ctx : rel_context) (pats : (int * bool) list) =
     List.map_i (fun i decl ->
         let (id, b, t) = to_named_tuple decl in
         let i', _ = lookup_named_i id.binder_name nctx in
-        CList.find_map_exn (fun (i'', hide) ->
+        CList.find_map (fun (i'', hide) ->
             if i'' == i' then Some (if hide then PHide i else PRel i)
             else None) pats)
       1 ctx'
-  in fst (rel_of_named_context evars ctx'), pats', pats''
+  in fst (rel_of_named_context ctx'), pats', pats''
 
 let push_rel_context_eos ctx env evars =
   if named_context env <> [] then
@@ -665,11 +655,11 @@ let push_rel_context_eos ctx env evars =
     in push_rel_context ctx env'
   else push_rel_context ctx env
 
-let split_at_eos env sigma ctx =
+let split_at_eos sigma ctx =
   List.split_when (fun decl ->
-      is_lglobal env sigma coq_end_of_section (get_named_type decl)) ctx
+      is_lglobal sigma coq_end_of_section (get_named_type decl)) ctx
 
-let pr_problem p env sigma { src_ctx = delta; map_inst = patcs }=
+let pr_problem p env sigma (delta, patcs, _) =
   let env' = push_rel_context delta env in
   let ctx = pr_context env sigma delta in
   Id.print p.program_id ++ str" " ++ pr_pats env' sigma patcs ++
@@ -684,7 +674,7 @@ let push_named_context = List.fold_right push_named
 let check_unused_clauses env sigma cl =
   let unused = List.filter (fun (_, (_, used)) -> used = 0) cl in
   match unused with
-  | (Pre_clause (loc, lhs, _) as cl, _) :: cls ->
+  | ((loc, lhs, _) as cl, _) :: cls ->
     user_err_loc (loc, str "Unused clause " ++ pr_preclause env sigma cl)
   | [] -> ()
 
@@ -709,9 +699,7 @@ let compute_rec_type context programs =
       user_err_loc (None, Pp.str "Mutual well-founded definitions are not supported");
     let p = List.hd programs in
     match p.program_rec with
-    | Some (WellFounded (_, _, id)) -> 
-      let nargs = Context.Rel.nhyps p.program_sign in
-      Some (Logical (nargs, id)) :: context
+    | Some (WellFounded (_, _, newinfo)) -> Some (Logical newinfo) :: context
     | _ -> assert false
   end
 
@@ -723,9 +711,7 @@ let print_program_info env sigma programs =
 let make_fix_proto env sigma ty =
   let relevance = Retyping.relevance_of_type env sigma ty in
   let _, fixproto = get_fresh sigma coq_fix_proto in
-  let r = Retyping.relevance_of_term env sigma fixproto in
-  let na = make_annot Anonymous r in
-  relevance, mkLetIn (na, fixproto, Retyping.get_type_of env sigma fixproto, lift 1 ty)
+    relevance, mkLetIn (anonR, fixproto, Retyping.get_type_of env sigma fixproto, lift 1 ty)
 
 let compute_fixdecls_data env evd ?data programs =
   let protos = List.map (fun p ->
@@ -747,7 +733,7 @@ let interp_arity env evd ~poly ~is_rec ~with_evars notations (((loc,i),udecl,rec
   let (arity, impls') =
     let ty = match t with
       | Some ty -> ty
-      | None -> CAst.make ?loc (Constrexpr.CHole None)
+      | None -> CAst.make ?loc (Constrexpr.CHole (None, Namegen.IntroAnonymous, None))
     in
     Equations_common.evd_comb1 (interp_type_evars_impls env' ?impls:None) evd ty
   in
@@ -784,9 +770,9 @@ let interp_arity env evd ~poly ~is_rec ~with_evars notations (((loc,i),udecl,rec
   let _ = if not with_evars then Pretyping.check_evars env !evd body in
   let program_orig_type = it_mkProd_or_LetIn arity sign in
   let program_sort =
-    let u = Retyping.get_sort_of env !evd program_orig_type in
+    let u = Sorts.univ_of_sort (Retyping.get_sort_of env !evd program_orig_type) in
     let sigma, sortl, sortu = nonalgebraic_universe_level_of_universe env !evd u in
-    evd := sigma; ESorts.kind sigma sortu
+    evd := sigma; sortu
   in
   let program_implicits = Impargs.compute_implicits_with_manual env !evd program_orig_type false impls in
   let () = evd := Evd.minimize_universes !evd in
@@ -841,10 +827,10 @@ let pats_of_sign sign =
   List.rev_map (fun decl ->
       DAst.make (PUVar (Name.get_id (Context.Rel.Declaration.get_name decl), Implicit))) sign
 
-let abstract_term_in_context env evars idx t map =
-  let before, after = CList.chop (pred idx) map.src_ctx in
+let abstract_term_in_context env evars idx t (g, p, d) =
+  let before, after = CList.chop (pred idx) g in
   let before' = subst_term_in_context evars (lift (- pred idx) t) before in
-  { src_ctx = before' @ after; map_inst = map.map_inst; tgt_ctx = map.tgt_ctx }
+  (before' @ after, p, d)
 
 let wf_fix_constr env evars sign arity sort carrier cterm crel =
   let sigma, tele, telety = Sigma_types.telescope_of_context env !evars sign in
@@ -867,12 +853,11 @@ let wf_fix_constr env evars sign arity sort carrier cterm crel =
   let sigma, fixty = Typing.type_of env !evars fix in
   let () = evars := sigma in
   let reds =
-    let flags = RedFlags.betaiotazeta in
+    let flags = CClosure.betaiotazeta in
     let csts =
       let ts = TransparentState.empty in
-      let tr_prj = Names.PRpred.add (Projection.repr (Lazy.force coq_pr1)) Names.PRpred.empty in
-      let tr_prj = Names.PRpred.add (Projection.repr (Lazy.force coq_pr2)) tr_prj in
-      let cst = Names.Cpred.empty in
+      let cst = Names.Cpred.add (Projection.constant (Lazy.force coq_pr1)) Names.Cpred.empty in
+      let cst = Names.Cpred.add (Projection.constant (Lazy.force coq_pr2)) cst in
       let add_ts cst t = Names.Cpred.add (Globnames.destConstRef (Lazy.force t)) cst in
       let tr_cst = List.fold_left add_ts cst
           [logic_tele_interp;
@@ -886,11 +871,11 @@ let wf_fix_constr env evars sign arity sort carrier cterm crel =
            logic_tele_forall;
            logic_tele_forall_pack;
            logic_tele_forall_unpack]
-      in { ts with TransparentState.tr_cst; TransparentState.tr_prj }
-    in RedFlags.red_add_transparent flags csts
+      in { ts with TransparentState.tr_cst }
+    in CClosure.RedFlags.red_add_transparent flags csts
   in
   let norm env =
-    let infos = Cbv.create_cbv_infos reds ~strong:true env !evars in
+    let infos = Cbv.create_cbv_infos reds env !evars in
     Cbv.cbv_norm infos
   in
   let fixty = norm env fixty in
@@ -905,11 +890,11 @@ let wf_fix_constr env evars sign arity sort carrier cterm crel =
   in
   let fix = norm env fix in
   let functional_type, full_functional_type =
-    let ctx, rest = Reductionops.whd_decompose_prod_n_assum env !evars (Context.Rel.nhyps sign) functional_type in
+    let ctx, rest = Reductionops.splay_prod_n env !evars (Context.Rel.nhyps sign) functional_type in
     match kind !evars (whd_all (push_rel_context ctx env) !evars rest) with
     | Prod (na, b, concl) ->
-      let ctx', rest = Reductionops.whd_decompose_prod_decls (push_rel_context ctx env) !evars b in
-      let infos = Cbv.create_cbv_infos reds ~strong:true (push_rel_context ctx env) !evars in
+      let ctx', rest = Reductionops.splay_prod_assum (push_rel_context ctx env) !evars b in
+      let infos = Cbv.create_cbv_infos reds (push_rel_context ctx env) !evars in
       let norm = Cbv.cbv_norm infos in
       let fn_type = it_mkProd_or_LetIn rest ctx' in
       let fn_type = norm fn_type in
@@ -928,12 +913,11 @@ let wf_fix_constr env evars sign arity sort carrier cterm crel =
 let wf_fix env evars subst sign arity sort term rel =
   let envsign = push_rel_context sign env in
   let sigma, cterm = interp_constr_evars envsign !evars term in
-  let na, carrier =
-    let r = Retyping.relevance_of_term envsign sigma cterm in
+  let carrier =
     let ty = Retyping.get_type_of envsign sigma cterm in
     let ty = nf_all envsign sigma ty in
     if noccur_between sigma 1 (length sign) ty then
-      make_annot Anonymous r, lift (- length sign) ty
+      lift (- length sign) ty
     else
       user_err_loc (Constrexpr_ops.constr_loc term,
                     str"The carrier type of the recursion order cannot depend on the arguments")
@@ -943,7 +927,7 @@ let wf_fix env evars subst sign arity sort term rel =
   let sigma, rsort = Evd.fresh_sort_in_family sigma (Lazy.force Equations_common.logic_sort) in
   let sigma, crel =
     let relty =
-      (mkProd (na, carrier, mkProd (na, lift 1 carrier, mkSort rsort)))
+      (mkProd (anonR, carrier, mkProd (anonR, lift 1 carrier, mkSort rsort)))
     in
     match rel with
     | Some rel -> interp_casted_constr_evars env sigma rel relty
@@ -1002,7 +986,7 @@ let compute_rec_data env evars data lets subst p =
     let rec_sign = p.program_sign @ lets in
     let lhs = decl :: rec_sign in
     let pats = PHide 1 :: lift_pats 1 (id_pats rec_sign) in
-    let rec_prob = { src_ctx = lhs; map_inst = pats; tgt_ctx = lhs } in
+    let rec_prob = (lhs, pats, lhs) in
     let rec_node =
       { wf_rec_term = fix;
         wf_rec_functional = None;
@@ -1025,8 +1009,7 @@ let compute_rec_data env evars data lets subst p =
 
 exception UnfaithfulSplit of (Loc.t option * Pp.t)
 
-let rename_domain env sigma bindings map = 
-  let { src_ctx = ctx; map_inst = p; tgt_ctx = ctx' } = map in
+let rename_domain env sigma bindings (ctx, p, ctx') = 
   let fn rel decl = 
     match Int.Map.find rel bindings with
     | exception Not_found -> decl
@@ -1055,14 +1038,13 @@ let loc_before loc loc' =
     end_ < end' || (end_ = end' && start <= start')
 
 let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int)) list) path
-    prob extpats lets ty =
-  let { src_ctx = ctx; map_inst = pats; tgt_ctx = ctx' } = prob in
+    (ctx,pats,ctx' as prob) extpats lets ty =
   if !Equations_common.debug then
     Feedback.msg_debug Pp.(str"Launching covering on "++ pr_preclauses env !evars (List.map fst clauses) ++
                            str " with problem " ++ pr_problem p env !evars prob ++
                            str " extpats " ++ pr_user_pats env !evars extpats);
   match clauses with
-  | (Pre_clause (loc, lhs, rhs), (idx, cnt) as clause) :: clauses' ->
+  | ((loc, lhs, rhs), (idx, cnt) as clause) :: clauses' ->
     if !Equations_common.debug then
       Feedback.msg_debug (str "Matching " ++ pr_user_pats env !evars (extpats @ lhs) ++ str " with " ++
                           pr_problem p env !evars prob);
@@ -1081,7 +1063,7 @@ let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int))
                 else Id.Map.add x (loc, gen, pat) acc
               else if data.flags.allow_aliases then acc
               else 
-                let env = push_rel_context prob.src_ctx env in
+                let env = push_rel_context (pi1 prob) env in
                 let loc, pat, pat' = 
                   if loc_before loc loc' then loc', pat, pat'
                   else loc, pat', pat
@@ -1160,9 +1142,9 @@ let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int))
                            (str "This pattern cannot be empty, it matches value " ++ fnl () ++
                             pr_pat env !evars c))
           | PRel i ->
-            match prove_empty (env,evars) prob.src_ctx i with
+            match prove_empty (env,evars) (pi1 prob) i with
             | Some (i, ctx, s) ->
-              Some (List.rev prev @ ((Pre_clause (loc, lhs, rhs),(idx, cnt+1)) :: clauses'),
+              Some (List.rev prev @ (((loc, lhs, rhs),(idx, cnt+1)) :: clauses'),
                     Compute (prob, [], ty, REmpty (i, s)))
             | None ->
               user_err_loc (loc,
@@ -1177,9 +1159,9 @@ let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int))
            | None ->
              user_err_loc
                (dummy_loc,
-                str"Clause " ++ pr_preclause env !evars (Pre_clause (loc, lhs, Some rhs)) ++
+                str"Clause " ++ pr_preclause env !evars (loc, lhs, Some rhs) ++
                 str" matched but its interpretation failed")
-           | Some s -> Some (List.rev prev @ (Pre_clause (loc,lhs,Some rhs),(idx, cnt+1)) :: clauses', s)))
+           | Some s -> Some (List.rev prev @ ((loc,lhs,Some rhs),(idx, cnt+1)) :: clauses', s)))
 
      | UnifFailure ->
        if !Equations_common.debug then Feedback.msg_debug (str "failed");
@@ -1190,26 +1172,26 @@ let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int))
        let blocks = blockers env (extpats @ lhs) prob in
        equations_debug (fun () ->
            str "blockers are: " ++
-           prlist_with_sep spc (pr_rel_name (push_rel_context prob.src_ctx env)) blocks);
+           prlist_with_sep spc (pr_rel_name (push_rel_context (pi1 prob) env)) blocks);
        let rec try_split acc vars =
          match vars with
          | [] -> None
          | var :: vars ->
            equations_debug (fun () -> str "trying next blocker " ++
-                                      pr_rel_name (push_rel_context prob.src_ctx env) var);
-           match split_var (env,evars) var prob.src_ctx with
+                                      pr_rel_name (push_rel_context (pi1 prob) env) var);
+           match split_var (env,evars) var (pi1 prob) with
            | Some (Splitted (var, newctx, s)) ->
              equations_debug (fun () ->
                  str "splitting succeded for " ++
-                 pr_rel_name (push_rel_context prob.src_ctx env) var);
-             let prob' = { src_ctx = newctx; map_inst = pats; tgt_ctx = ctx' } in
+                 pr_rel_name (push_rel_context (pi1 prob) env) var);
+             let prob' = (newctx, pats, ctx') in
              let coverrec clauses s =
                covering_aux env evars p data []
                  clauses path
                  (compose_subst env ~sigma:!evars s prob')
                  extpats
-                 (specialize_rel_context !evars s.map_inst lets)
-                 (specialize_constr !evars s.map_inst ty)
+                 (specialize_rel_context !evars (pi2 s) lets)
+                 (specialize_constr !evars (pi2 s) ty)
              in
              (try
                 let rec_call clauses x =
@@ -1237,7 +1219,7 @@ let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int))
            | Some (CannotSplit _) as x ->
              equations_debug (fun () ->
                  str "splitting failed for " ++
-                 pr_rel_name (push_rel_context prob.src_ctx env) var);
+                 pr_rel_name (push_rel_context (pi1 prob) env) var);
              let acc = match acc with
                | None -> x
                | _ -> acc
@@ -1256,7 +1238,7 @@ let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int))
        | None -> None))
   | [] -> (* Every clause failed for the problem, it's either uninhabited or
              the clauses are not exhaustive *)
-    match find_empty (env,evars) prob.src_ctx with
+    match find_empty (env,evars) (pi1 prob) with
     | Some (i, ctx, s) ->
       Some (List.rev prev @ clauses, (* Split (prob, i, ty, s)) *)
             Compute (prob, [], ty, REmpty (i, s)))
@@ -1265,10 +1247,9 @@ let rec covering_aux env evars p data prev (clauses : (pre_clause * (int * int))
         (str "Non-exhaustive pattern-matching, no clause found for:" ++ fnl () ++
          pr_problem p env !evars prob))
 
-and interp_clause env evars p data prev clauses' path prob
+and interp_clause env evars p data prev clauses' path (ctx,pats,ctx' as prob)
     extpats lets ty
     ((loc,lhs,rhs), used) (s, uinnacs, innacs) =
-  let { src_ctx = ctx; map_inst = pats; tgt_ctx = ctx' } = prob in
   let env' = push_rel_context_eos ctx env evars in
   let get_var loc i s =
     match assoc i s with
@@ -1283,7 +1264,7 @@ and interp_clause env evars p data prev clauses' path prob
       let userc, usercty = interp_constr_in_rhs env ctx evars data None s lets (GlobConstr user) in
       match t with
       | PInac t ->
-        begin match Evarconv.unify env' !evars Conversion.CONV userc t with
+        begin match Evarconv.unify env' !evars Reduction.CONV userc t with
           | evars' -> evars := evars'
           | exception Pretype_errors.PretypeError (env, sigma, e) ->
             DAst.with_loc_val (fun ?loc _ ->
@@ -1339,7 +1320,7 @@ and interp_clause env evars p data prev clauses' path prob
     Some res
 
   | Empty (loc,i) ->
-    (match prove_empty (env, evars) prob.src_ctx (get_var loc i s) with
+    (match prove_empty (env, evars) (pi1 prob) (get_var loc i s) with
      | None -> user_err_loc (loc, str"Cannot show that " ++ Id.print i ++ str"'s type is empty")
      | Some (i, ctx, s) ->
        Some (Compute (prob, [], ty, REmpty (i, s))))
@@ -1358,7 +1339,7 @@ and interp_clause env evars p data prev clauses' path prob
     let newctx, pats', pats'' = instance_of_pats env !evars ctx vars in
     (* revctx is a variable substitution from a reordered context to the
        current context *)
-    let revctx = check_ctx_map env !evars { src_ctx = newctx; map_inst = pats'; tgt_ctx = ctx } in
+    let revctx = check_ctx_map env !evars (newctx, pats', ctx) in
     let idref = Namegen.next_ident_away (Id.of_string "refine") (Id.Set.of_list (ids_of_rel_context newctx)) in
     let refterm (* in newctx *) = mapping_constr !evars revctx cconstr in
     let refty = mapping_constr !evars revctx cty in
@@ -1390,42 +1371,40 @@ and interp_clause env evars p data prev clauses' path prob
     let refterm (* in extnewctx *) = lift 1 refterm in
     equations_debug Pp.(fun () -> str" Strenghtening variable decl: " ++ pr_context_map env !evars tytop);
     equations_debug Pp.(fun () -> str" Strenghtening variable decl inv: " ++ pr_context_map env !evars tytopinv);
-    let cmap, strinv = new_strengthen env !evars tytop.src_ctx 1 refterm in
+    let cmap, strinv = new_strengthen env !evars (pi1 tytop) 1 refterm in
     let cmap = compose_subst env ~sigma:!evars cmap tytop in
     let strinv = compose_subst env ~sigma:!evars tytopinv strinv in
-    let strinv_map = List.map_i (fun i -> function (PRel j) -> i, j | _ -> assert false) 1 strinv.map_inst in
+    let strinv_map = List.map_i (fun i -> function (PRel j) -> i, j | _ -> assert false) 1 (pi2 strinv) in
     equations_debug Pp.(fun () -> str" Strenghtening: " ++ pr_context_map env !evars cmap);
     equations_debug Pp.(fun () -> str" Strenghtening inverse: " ++ pr_context_map env !evars strinv);
-    let idx_of_refined = destPRel (CList.hd cmap.map_inst) in
+    let idx_of_refined = destPRel (CList.hd (pi2 cmap)) in
     equations_debug Pp.(fun () -> str" idx_of_refined: " ++ int idx_of_refined);
     let cmap =
       abstract_term_in_context env !evars idx_of_refined (mapping_constr !evars cmap refterm) cmap
     in
     equations_debug Pp.(fun () -> str" Strenghtening + abstraction: " ++ pr_context_map env !evars cmap);
     let newprob_to_lhs =
-      let inst_refctx = set_in_ctx idx_of_refined (mapping_constr !evars cmap refterm) cmap.src_ctx in
-      let str_to_new = {
-        src_ctx = inst_refctx;
-        map_inst = specialize_pats !evars cmap.map_inst (lift_pats 1 pats');
-        tgt_ctx = newctx;
-      } in
+      let inst_refctx = set_in_ctx idx_of_refined (mapping_constr !evars cmap refterm) (pi1 cmap) in
+      let str_to_new =
+        inst_refctx, (specialize_pats !evars (pi2 cmap) (lift_pats 1 pats')), newctx
+      in
       equations_debug Pp.(fun () -> str" Strenghtening + abstraction + instantiation: " ++ pr_context_map env !evars str_to_new);
       compose_subst env ~sigma:!evars str_to_new revctx
     in	
     equations_debug Pp.(fun () -> str" Strenghtening + abstraction + instantiation: " ++ pr_context_map env !evars newprob_to_lhs);
 
     let newprob = 
-      let ctx = cmap.src_ctx in
+      let ctx = pi1 cmap in
       let pats = 
         rev_map (fun c -> 
             let idx = destRel !evars c in
             (* find out if idx in ctx should be hidden depending
                on its use in newprob_to_lhs *)
             if List.exists (function PHide idx' -> idx == idx' | _ -> false)
-                newprob_to_lhs.map_inst then
+                (pi2 newprob_to_lhs) then
               PHide idx
             else PRel idx) (rels_of_ctx ctx) in
-      { src_ctx = ctx; map_inst = pats; tgt_ctx = ctx }
+      (ctx, pats, ctx)
     in
     let newty =
       let env' = push_rel_context extnewctx env in
@@ -1443,7 +1422,7 @@ and interp_clause env evars p data prev clauses' path prob
       match cs with
       | [] -> cls
       | _ :: _ ->
-        [Pre_clause (loc, lhs @ [DAst.make ?loc (PUVar (idref, Generated))], Some (Refine (cs, cls)))]
+        [loc, lhs @ [DAst.make ?loc (PUVar (idref, Generated))], Some (Refine (cs, cls))]
     in
     let rec cls' n cls =
       let next_unknown =
@@ -1451,7 +1430,7 @@ and interp_clause env evars p data prev clauses' path prob
         let i = ref (-1) in fun () ->
           incr i; add_suffix str (string_of_int !i)
       in
-      List.map_filter (fun (Pre_clause (loc, lhs, rhs)) ->
+      List.map_filter (fun (loc, lhs, rhs) ->
         let oldpats, newpats = List.chop (List.length lhs - n) lhs in
         let newref, nextrefs =
           match newpats with hd :: tl -> hd, tl | [] -> assert false
@@ -1476,7 +1455,7 @@ and interp_clause env evars p data prev clauses' path prob
             | Some (Refine (cs', cls)) -> Some (Refine (cs', cls' (List.length cs' + n) cls))
             | _ -> rhs
           in
-          Some (Pre_clause (loc, rev newlhs @ nextrefs, newrhs))
+          Some (loc, rev newlhs @ nextrefs, newrhs)
         | _ ->
           CErrors.user_err ?loc
             (str "Non-matching clause in with subprogram:" ++ fnl () ++ int n ++
@@ -1498,7 +1477,7 @@ and interp_clause env evars p data prev clauses' path prob
     let strength_app = List.map_filter (fun t ->
         if isRel !evars t then
           let i = destRel !evars t in
-          let decl = List.nth prob.src_ctx (pred i) in
+          let decl = List.nth (pi1 prob) (pred i) in
           if Context.Rel.Declaration.is_local_def decl then None
           else Some t
         else Some t) strength_app
@@ -1509,7 +1488,7 @@ and interp_clause env evars p data prev clauses' path prob
       let _, ctxs, _ = lets_of_ctx env ctx evars s in
       let newlets = (lift_rel_context (succ letslen) ctxs)
                     @ (lift_rel_context 1 lets)
-      in specialize_rel_context !evars cmap.map_inst newlets
+      in specialize_rel_context !evars (pi2 cmap) newlets
     in
     let clauses' = List.mapi (fun i x -> x, (succ i, 0)) cls' in
     match covering_aux env evars p data [] clauses' path' newprob [] lets' newty with
@@ -1520,11 +1499,11 @@ and interp_clause env evars p data prev clauses' path prob
          str "And clauses: " ++ pr_preclauses env !evars cls')
     | Some (clauses, s) ->
       let () = check_unused_clauses env !evars clauses in
-      let term, _ = term_of_tree env evars (ESorts.make p.program_sort) s in
+      let term, _ = term_of_tree env evars p.program_sort s in
       let info =
         { refined_obj = (idref, cconstr, cty);
           refined_rettyp = ty;
-          refined_arg = refine_arg idx_of_refined cmap.src_ctx;
+          refined_arg = refine_arg idx_of_refined (pi1 cmap);
           refined_path = path';
           refined_term = term;
           refined_filter = None;
@@ -1542,14 +1521,14 @@ and interp_clause env evars p data prev clauses' path prob
 
 and interp_wheres env0 ctx evars path data s lets
     (ctx, envctx, liftn, subst)
-    (w : (pre_prototype * pre_equation list) list * Vernacexpr.notation_declaration list) =
+    (w : (pre_prototype * pre_equation list) list * Vernacexpr.decl_notation list) =
   let notations = snd w in
   let aux (data,lets,nlets,coverings,env)
       (((loc,id),udecl,nested,b,t,reca),clauses as eqs) =
 
     let is_rec = is_recursive id ([eqs], notations) in
     let p = interp_arity env evars ~poly:false ~is_rec ~with_evars:true notations eqs in
-    let clauses = List.map (interp_eqn env !evars (data.notations @ notations) p ~avoid:Id.Set.empty) clauses in
+    let clauses = List.map (interp_eqn env (data.notations @ notations) p ~avoid:Id.Set.empty) clauses in
     let sigma, p = adjust_sign_arity env !evars p clauses in
     let () = evars := sigma in
 
@@ -1624,7 +1603,7 @@ and covering ?(check_unused=true) env evars p data (clauses : pre_clause list)
        pr_problem p env !evars prob)
 
 let program_covering env evd data p clauses =
-  let clauses = List.map (interp_eqn (push_rel_context data.fixdecls env) !evd
+  let clauses = List.map (interp_eqn (push_rel_context data.fixdecls env) 
     data.notations p ~avoid:Id.Set.empty) clauses in
   let sigma, p = adjust_sign_arity env !evd p clauses in
   let () = evd := sigma in
