@@ -78,8 +78,9 @@ let needs_generalization gl id =
           let f, args = decompose_app sigma t in
             f, args, true, id, oldid
   in
-    if Array.is_empty args then false
+    if args = [] then false
     else
+      let args = Array.of_list args in
       let f', args' = decompose_indapp sigma f args in
       let parvars = ids_of_constr ~all:true sigma Id.Set.empty f' in
         if not (linear sigma parvars args') then true
@@ -90,7 +91,6 @@ let dependent_pattern ?(pattern_term=true) c =
   let open Tacmach in
   Proofview.Goal.enter (fun gl ->
   let sigma = Proofview.Goal.sigma gl in
-  let cr = Retyping.relevance_of_term (pf_env gl) sigma c in
   let cty = Retyping.get_type_of (pf_env gl) sigma c in
   let deps =
     match kind sigma cty with
@@ -104,21 +104,20 @@ let dependent_pattern ?(pattern_term=true) c =
     | _ -> pf_get_new_id (Id.of_string (hdchar (pf_env gl) (project gl) c)) gl
   in
   let env = pf_env gl in
-  let mklambda (ty, sigma) (c, id, r, cty) =
+  let mklambda (ty, sigma) (c, id, cty) =
     let conclvar, sigma =
       Find_subterm.subst_closed_term_occ env sigma
         (Locus.AtOccs Locus.AllOccurrences) c ty
     in
-      mkNamedLambda sigma (make_annot id r) cty conclvar, sigma
+      mkNamedLambda (annotR id) cty conclvar, sigma
   in
   let subst =
-    let map c = (c, varname c, Retyping.relevance_of_term env sigma c, Retyping.get_type_of env sigma c) in
-    let deps = List.rev_map map deps in
-    if pattern_term then (c, varname c, cr, cty) :: deps
-    else deps
+    let deps = List.rev_map (fun c -> (c, varname c, pf_get_type_of gl c)) deps in
+      if pattern_term then (c, varname c, cty) :: deps
+      else deps
   in
   let concllda, evd = List.fold_left mklambda (pf_concl gl, project gl) subst in
-  let conclapp = applistc concllda (List.rev_map (fun (c, _, _, _) -> c) subst) in
+  let conclapp = applistc concllda (List.rev_map pi1 subst) in
     convert_concl ~cast:false ~check:false conclapp DEFAULTcast)
 
 let annot_of_context ctx =
@@ -127,8 +126,8 @@ let annot_of_context ctx =
 let depcase ~poly ((mind, i as ind), u) =
   let indid = Nametab.basename_of_global (GlobRef.IndRef ind) in
   let mindb, oneind = Global.lookup_inductive ind in
-  let relevance = ERelevance.make oneind.mind_relevance in
-  let indna x = make_annot x relevance in
+  let relevance = oneind.mind_relevance in
+  let annotR x = make_annot x relevance in
   let ctx = oneind.mind_arity_ctxt in
   let nparams = mindb.mind_nparams in
   let ctx = List.map of_rel_decl ctx in
@@ -136,28 +135,27 @@ let depcase ~poly ((mind, i as ind), u) =
   let nargs = List.length args in
   let indapp = mkApp (mkIndU (ind,u), extended_rel_vect 0 ctx) in
   let evd = ref (Evd.from_env (Global.env())) in
-  let s = evd_comb0 (Evd.new_sort_variable Evd.univ_flexible) evd in
-  let pred = it_mkProd_or_LetIn (mkSort s)
-    (make_assum (indna Anonymous) indapp :: args)
+  let pred = it_mkProd_or_LetIn (evd_comb0 Evarutil.new_Type evd)
+    (make_assum anonR indapp :: args)
   in
   let nconstrs = Array.length oneind.mind_nf_lc in
   let mkbody i (ctx, ty) =
     let args = Context.Rel.instance mkRel 0 ctx in
-    annot_of_context (EConstr.of_rel_context ctx), mkApp (mkRel (1 + nconstrs + List.length ctx - i), args)
+    annot_of_context ctx, mkApp (mkRel (1 + nconstrs + List.length ctx - i), args)
   in
   let bodies = Array.mapi mkbody oneind.mind_nf_lc in
   let branches =
     Array.map2_i (fun i id (ctx, cty) ->
       let cty = Term.it_mkProd_or_LetIn cty ctx in
-      let substcty = Vars.subst_instance_constr u (of_constr cty) in
-      let (args, arity) = decompose_prod_decls !evd substcty in
+      let substcty = Vars.subst_instance_constr (EInstance.kind !evd u) (of_constr cty) in
+      let (args, arity) = decompose_prod_assum !evd substcty in
       let _, indices = decompose_app !evd arity in
-      let _, indices = Array.chop nparams indices in
+      let _, indices = List.chop nparams indices in
       let ncargs = List.length args - nparams in
       let realargs, pars = List.chop ncargs args in
       let realargs = lift_rel_context (i + 1) realargs in
       let arity = applistc (mkRel (ncargs + i + 1))
-        (Array.to_list indices @ [mkApp (mkConstructU ((ind, succ i), u),
+        (indices @ [mkApp (mkConstructU ((ind, succ i), u),
                           Array.append (extended_rel_vect (ncargs + i + 1) params)
                             (extended_rel_vect 0 realargs))])
       in
@@ -165,29 +163,28 @@ let depcase ~poly ((mind, i as ind), u) =
         (make_assum (nameR (Id.of_string ("P" ^ string_of_int i))) br))
       oneind.mind_consnames oneind.mind_nf_lc
   in
-  let ci = make_case_info (Global.env ()) ind RegularStyle in
+  let ci = make_case_info (Global.env ()) ind relevance RegularStyle in
   let obj i =
     mkApp (mkIndU (ind,u),
           (Array.append (extended_rel_vect (nargs + nconstrs + i) params)
               (extended_rel_vect 0 args)))
   in
-  let ctxpred = make_assum (indna Anonymous) (obj (2 + nargs)) :: args in
+  let ctxpred = make_assum (annotR Anonymous) (obj (2 + nargs)) :: args in
   let app = mkApp (mkRel (nargs + nconstrs + 3),
                   (extended_rel_vect 0 ctxpred))
   in
   let paramsinst = extended_rel_vect (2 + nargs + nconstrs) params in
   let ty = (annot_of_context ctxpred, app) in
-  let case = mkCase (ci, EInstance.empty, paramsinst, (ty, relevance), NoInvert, mkRel 1, bodies) in
+  let case = mkCase (ci, EInstance.empty, paramsinst, ty, NoInvert, mkRel 1, bodies) in
   let xty = obj 1 in
   let xid = Namegen.named_hd (Global.env ()) !evd xty Anonymous in
   let body =
     let len = 1 (* P *) + Array.length branches in
     it_mkLambda_or_LetIn case
-      (make_assum (indna xid) (lift len indapp)
+      (make_assum (annotR xid) (lift len indapp)
         :: ((Array.rev_to_list branches)
-            @ (make_assum (make_annot (Name (Id.of_string "P")) (Retyping.relevance_of_sort s)) pred :: ctx)))
+            @ (make_assum (nameR (Id.of_string "P")) pred :: ctx)))
   in
-  let () = evd := Evd.minimize_universes !evd in
   let univs = Evd.univ_entry ~poly !evd in
   let ce = Declare.definition_entry ~univs (EConstr.to_constr !evd body) in
   let kn =
@@ -225,8 +222,7 @@ let pattern_call ?(pattern_term=true) c =
   Proofview.Goal.enter (fun gl ->
   let env = pf_env gl in
   let sigma = project gl in
-  let cr = Retyping.relevance_of_term env sigma c in
-  let cty = Retyping.get_type_of env sigma c in
+  let cty = pf_get_type_of gl c in
   let ids = Id.Set.of_list (ids_of_named_context (Proofview.Goal.hyps gl)) in
   let deps =
     match kind sigma c with
@@ -241,7 +237,7 @@ let pattern_call ?(pattern_term=true) c =
   let mklambda ty (c, id, cty) =
     let conclvar, _ = Find_subterm.subst_closed_term_occ env (project gl)
       (Locus.AtOccs Locus.AllOccurrences) c ty in
-      mkNamedLambda sigma (make_annot id cr) cty conclvar
+      mkNamedLambda (annotR id) cty conclvar
   in
   let subst =
     let deps = List.rev_map (fun c -> (c, varname c, pf_get_type_of gl c)) deps in
@@ -260,10 +256,7 @@ let destPolyRef sigma c =
   | Construct (cstr, u) -> ConstructRef cstr, u
   | _ -> raise (Invalid_argument "destPolyRef")
 
-(** Compare up-to variables in v, skipping parameters of inductive constructors. 
-    [t] is closed  
-  *)
-
+(** Compare up-to variables in v, skipping parameters of inductive constructors. *)
 let rec compare_upto_variables sigma t v =
   if (isVar sigma v || isRel sigma v) then true
   else
@@ -286,7 +279,7 @@ let whd_head env sigma t =
     mkApp (eq, Array.map (Tacred.whd_simpl env sigma) args)
   | _ -> t
 
-let specialize_eqs ?with_block id =
+let specialize_eqs ~with_block id =
   Proofview.Goal.enter begin fun gl ->
   let open Tacticals in
   let open Tacmach in
@@ -294,23 +287,23 @@ let specialize_eqs ?with_block id =
   let ty = pf_get_hyp_typ id gl in
   let evars = ref (project gl) in
   let unif env ctx evars c1 c2 =
-    match Evarconv.unify env !evars Conversion.CONV (it_mkLambda_or_subst env c1 ctx) (it_mkLambda_or_subst env c2 ctx) with
+    match Evarconv.unify env !evars Reduction.CONV (it_mkLambda_or_subst env c1 ctx) (it_mkLambda_or_subst env c2 ctx) with
     | exception Evarconv.UnableToUnify _ -> false
     | evm -> evars := evm; true
   in
-  let rec aux block_count in_block in_eqs ctx subst acc ty =
+  let rec aux in_block in_eqs ctx subst acc ty =
     match kind !evars ty with
     | LetIn (na, b, t, ty) ->
-      if is_global env !evars (Lazy.force coq_block) b then
-        if with_block = None then aux block_count in_block in_eqs ctx subst acc (subst1 mkProp ty)
-        else if (in_block || in_eqs) && Int.equal block_count 0 then acc, in_eqs, ctx, subst, (subst1 mkProp ty)
-        else aux (block_count - 1) true in_eqs ctx subst acc (subst1 mkProp ty)
+      if is_global !evars (Lazy.force coq_block) b then
+        if not with_block then aux true in_eqs ctx subst acc (subst1 mkProp ty)
+        else if in_block then acc, in_eqs, ctx, subst, (subst1 mkProp ty)
+        else aux true in_eqs ctx subst acc (subst1 mkProp ty)
       else if not in_block then
-        aux block_count in_block in_eqs (make_def na (Some b) t :: ctx) subst (lift 1 acc) ty
+        aux in_block in_eqs (make_def na (Some b) t :: ctx) subst acc ty
       else
-        aux block_count in_block in_eqs ctx (make_def na (Some b) t :: subst) acc ty
+        aux in_block in_eqs ctx (make_def na (Some b) t :: subst) acc ty
     | Prod (na, t, b) when not in_block ->
-      aux block_count false in_eqs (make_def na None t :: ctx) subst (mkApp (lift 1 acc, [| mkRel 1 |])) b
+      aux false in_eqs (make_def na None t :: ctx) subst (mkApp (lift 1 acc, [| mkRel 1 |])) b
     | Prod (na, t, b) ->
       let env' = push_rel_context ctx env in
       let env' = push_rel_context subst env' in
@@ -321,7 +314,7 @@ let specialize_eqs ?with_block id =
         str " to " ++ Printer.pr_econstr_env env' !evars t'); *)
       (match kind !evars t' with
        | App (eq, [| eqty; x; y |]) when
-           (is_global env !evars (Lazy.force logic_eq_type) eq &&
+           (is_global !evars (Lazy.force logic_eq_type) eq &&
             (noccur_between !evars 1 (List.length subst) x ||
              noccur_between !evars 1 (List.length subst) y)) ->
          let _, u = destPolyRef !evars eq in
@@ -329,22 +322,21 @@ let specialize_eqs ?with_block id =
            else y, x in
          let eqr = constr_of_global_univ !evars (Lazy.force logic_eq_refl, u) in
          let p = mkApp (eqr, [| eqty; c |]) in
-         if ((Option.equal Int.equal with_block (Some 2) && Int.equal block_count 0) || 
-              compare_upto_variables !evars c o) &&
+         if (compare_upto_variables !evars c o) &&
             unif (push_rel_context ctx env) subst evars o c then
-           aux block_count in_block true ctx subst (mkApp (acc, [| p |])) (subst1 p b)
+           aux in_block true ctx subst (mkApp (acc, [| p |])) (subst1 p b)
          else acc, in_eqs, ctx, subst, ty
        | _ ->
          if in_eqs then
            (* aux in_block false ctx (make_def na None t :: subst) (mkApp (lift 1 acc, [| mkRel 1 |])) b *)
            acc, in_eqs, ctx, subst, ty
          else
-           let e = evd_comb1 (Evarutil.new_evar env') evars t in
-           aux block_count in_block false ctx (make_def na (Some e) t :: subst) (mkApp (lift 1 acc, [| mkRel 1 |])) b)
+           let e = evd_comb1 (Evarutil.new_evar (push_rel_context ctx env))
+               evars (it_mkLambda_or_subst env t subst) in
+           aux in_block false ctx (make_def na (Some e) t :: subst) (mkApp (lift 1 acc, [| mkRel 1 |])) b)
     | t -> acc, in_eqs, ctx, subst, ty
   in
-  let acc, worked, ctx, subst, ty = aux (match with_block with None -> 0 | Some n -> n) 
-    (match with_block with None -> true | Some _ -> false) false [] [] (mkVar id) ty in
+  let acc, worked, ctx, subst, ty = aux (if with_block then false else true) false [] [] (mkVar id) ty in
   let subst' = nf_rel_context_evar !evars subst in
   let subst'' = List.map (fun decl ->
     let (n,b,t) = to_tuple decl in
@@ -359,8 +351,9 @@ let specialize_eqs ?with_block id =
   let ty = Evarutil.nf_evar !evars ty in
   let acc = Evarutil.nf_evar !evars acc in
     if worked then
-      assert_replacing id acc ty
-    else tclFAIL (str "Nothing to do in hypothesis " ++ Id.print id ++
+      tclTHENFIRST (Tactics.assert_before_replacing id ty)
+        (exact_no_check acc)
+    else tclFAIL 0 (str "Nothing to do in hypothesis " ++ Id.print id ++
                     Printer.pr_econstr_env env !evars ty
                    )
   end
@@ -369,12 +362,12 @@ exception Specialize
 
 open Proofview.Notations
 
-let specialize_eqs ?with_block id =
+let specialize_eqs ~with_block id =
   let open Tacticals in
   Proofview.Goal.enter begin fun gl ->
   Proofview.tclORELSE (clear [id] <*> Proofview.tclZERO Specialize) begin function
-  | (Specialize, _) -> specialize_eqs ?with_block id
-  | e -> tclFAIL (str "Specialization not allowed on dependent hypotheses")
+  | (Specialize, _) -> specialize_eqs ~with_block id
+  | e -> tclFAIL 0 (str "Specialization not allowed on dependent hypotheses")
   end
   end
 
@@ -384,7 +377,7 @@ let dependent_elim_tac ?patterns id : unit Proofview.tactic =
     let env = Proofview.Goal.env gl in
     let concl = Proofview.Goal.concl gl in
     let sigma = Proofview.Goal.sigma gl in
-    let sort = Retyping.get_sort_of env sigma concl in
+    let sort = Sorts.univ_of_sort (Retyping.get_sort_of env sigma concl) in
     let env = Environ.reset_context env in
     let hyps = Proofview.Goal.hyps gl in
     let default_loc, id = id in
@@ -407,15 +400,15 @@ let dependent_elim_tac ?patterns id : unit Proofview.tactic =
     end >>= fun rel ->
 
     (* We want to work in a [rel_context], not a [named_context]. *)
-    let ctx, subst = Equations_common.rel_of_named_context sigma loc_hyps in
+    let ctx, subst = Equations_common.rel_of_named_context loc_hyps in
     let _, rev_subst, _ =
       let err () = assert false in
       Equations_common.named_of_rel_context ~keeplets:true err ctx in
     (* We also need to convert the goal for it to be well-typed in
      * the [rel_context]. *)
-    let ty = Vars.subst_vars sigma subst concl in
+    let ty = Vars.subst_vars subst concl in
     let rhs =
-      let prog = Constrexpr.CHole None in
+      let prog = Constrexpr.CHole (None, Namegen.IntroAnonymous, None) in
         Syntax.Program (Syntax.ConstrExpr (CAst.make prog), ([], []))
     in
     begin match patterns with
@@ -428,7 +421,7 @@ let dependent_elim_tac ?patterns id : unit Proofview.tactic =
         | Some (Covering.Splitted (_, newctx, brs)) ->
             let brs = Option.List.flatten (Array.to_list brs) in
             let clauses_lhs = List.map Context_map.context_map_to_lhs brs in
-            let clauses = List.map (fun lhs -> Syntax.Pre_clause (default_loc, lhs, Some rhs)) clauses_lhs in
+            let clauses = List.map (fun lhs -> (default_loc, lhs, Some rhs)) clauses_lhs in
               Proofview.tclUNIT clauses
         end
     | Some patterns ->
@@ -441,7 +434,7 @@ let dependent_elim_tac ?patterns id : unit Proofview.tactic =
                 if Names.Id.equal decl_id id then DAst.make ?loc pat
                 else DAst.make Syntax.(PUVar (decl_id, Generated))) loc_hyps
             in
-            Syntax.Pre_clause (loc, lhs, Some rhs))
+              (loc, lhs, Some rhs))
         in Proofview.tclUNIT (List.map make_clause patterns)
     end >>= fun clauses ->
     if !debug then
@@ -463,7 +456,7 @@ let dependent_elim_tac ?patterns id : unit Proofview.tactic =
     let program_orig_type = it_mkProd_or_LetIn ty ctx in
     let p = Syntax.{program_loc = default_loc;
                     program_id = Names.Id.of_string "dummy";
-                    program_orig_type; program_sort = (ESorts.kind sigma sort);
+                    program_orig_type; program_sort = sort;
                     program_impls = [];
                     program_implicits = [];
                     program_rec = None;
@@ -495,13 +488,12 @@ let dependent_elim_tac ?patterns id : unit Proofview.tactic =
 let dependent_elim_tac_expr ?patterns id : unit Proofview.tactic =
   Proofview.Goal.enter begin fun gl ->
     let env = Proofview.Goal.env gl in
-    let sigma = Proofview.Goal.sigma gl in
     (* Interpret each pattern to then produce clauses. *)
     let patterns =
       match patterns with
       | None -> None
       | Some p ->
         let avoid = Syntax.ids_of_pats None p in
-        Some (List.map (fun x -> List.hd (snd (Syntax.interp_pat env sigma [] ~avoid None x))) p)
+        Some (List.map (fun x -> List.hd (snd (Syntax.interp_pat env [] ~avoid None x))) p)
     in dependent_elim_tac ?patterns id
   end
