@@ -40,64 +40,55 @@ let equations_transparent = ref false
 let equations_derive_equations = ref true
 let equations_derive_eliminator = ref true
 
-let depr_with_k = Deprecation.make ~since:"equations v1.2" ()
-
-let () = Goptions.declare_bool_option {
-  Goptions.optdepr  = Some depr_with_k;
-  Goptions.optstage = Interp;
+let _ = Goptions.declare_bool_option {
+  Goptions.optdepr  = true;
   Goptions.optkey   = ["Equations"; "WithK"];
   Goptions.optread  = (fun () -> false);
   Goptions.optwrite = (fun b ->
       if b then
         CErrors.user_err (str"DEPRECATED. Use flag [Equations With UIP] and introduce \
-                              an axiom [forall A, Equations.Classes.UIP A] \
+                              an axiomn [forall A, Equations.Classes.UIP A] \
                               as a type class instance using [Existing Instance] instead.")
       else simplify_withUIP := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = Some depr_with_k;
-  Goptions.optstage = Interp;
+  Goptions.optdepr  = true;
   Goptions.optkey   = ["Equations"; "WithKDec"];
   Goptions.optread  = (fun () -> !simplify_withUIP);
   Goptions.optwrite = (fun b -> simplify_withUIP := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = None;
-  Goptions.optstage = Interp;
+  Goptions.optdepr  = false;
   Goptions.optkey   = ["Equations"; "With"; "UIP"];
   Goptions.optread  = (fun () -> !simplify_withUIP);
   Goptions.optwrite = (fun b -> simplify_withUIP := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = None;
-  Goptions.optstage = Interp;
+  Goptions.optdepr  = false;
   Goptions.optkey   = ["Equations"; "Transparent"];
   Goptions.optread  = (fun () -> !equations_transparent);
   Goptions.optwrite = (fun b -> equations_transparent := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = None;
-  Goptions.optstage = Interp;
+  Goptions.optdepr  = false;
   Goptions.optkey   = ["Equations"; "With"; "Funext"];
   Goptions.optread  = (fun () -> !equations_with_funext);
   Goptions.optwrite = (fun b -> equations_with_funext := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = None;
-  Goptions.optstage = Interp;
+  Goptions.optdepr  = false;
   Goptions.optkey   = ["Equations"; "Derive"; "Equations"];
   Goptions.optread  = (fun () -> !equations_derive_equations);
   Goptions.optwrite = (fun b -> equations_derive_equations := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = None;
-  Goptions.optstage = Interp;
+  Goptions.optdepr  = false;
   Goptions.optkey   = ["Equations"; "Derive"; "Eliminator"];
   Goptions.optread  = (fun () -> !equations_derive_eliminator);
   Goptions.optwrite = (fun b -> equations_derive_eliminator := b)
@@ -108,8 +99,7 @@ let _ = Goptions.declare_bool_option {
 let debug = ref false
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = None;
-  Goptions.optstage = Interp;
+  Goptions.optdepr  = false;
   Goptions.optkey   = ["Equations"; "Debug"];
   Goptions.optread  = (fun () -> !debug);
   Goptions.optwrite = (fun b -> debug := b)
@@ -153,7 +143,7 @@ let typecheck_rel_context env evd ctx =
   in ()
   with e ->
     Printf.eprintf "Exception while typechecking context %s : %s\n"
-                   (Pp.string_of_ppcmds (Termops.Internal.print_rel_context (EConstr.push_rel_context ctx env) evd))
+                   (Pp.string_of_ppcmds (Internal.print_rel_context (EConstr.push_rel_context ctx env)))
                    (Printexc.to_string e);
     raise e
 
@@ -220,15 +210,6 @@ let e_type_of env evd t =
   let evm, t = Typing.type_of ~refresh:false env !evd t in
   evd := evm; t
 
-let collapse_term_qualities uctx c =
-  let nf_evar _ = None in
-  let nf_qvar q = match UState.nf_qvar uctx q with
-    | QConstant _ as q -> q
-    | QVar q -> (* hack *) QConstant QType
-  in
-  let nf_univ _ = None in
-  UnivSubst.nf_evars_and_universes_opt_subst nf_evar nf_qvar nf_univ c
-
 let make_definition ?opaque ?(poly=false) evm ?types b =
   let env = Global.env () in
   let evm = match types with
@@ -239,14 +220,9 @@ let make_definition ?opaque ?(poly=false) evm ?types b =
   in
   let evm = Evd.minimize_universes evm in
   let evm0 = evm in
-  let to_constr c = collapse_term_qualities (Evd.evar_universe_context evm) (EConstr.to_constr evm c) in
-  let body = to_constr b in
-  let typ = Option.map to_constr types in
+  let body = EConstr.to_constr evm b and typ = Option.map (EConstr.to_constr evm) types in
   let used = Vars.universes_of_constr body in
-  let used' = match typ with
-    | None -> Univ.Level.Set.empty
-    | Some typ -> Vars.universes_of_constr typ
-  in
+  let used' = Option.cata Vars.universes_of_constr Univ.Level.Set.empty typ in
   let used = Univ.Level.Set.union used used' in
   let evm = Evd.restrict_universe_context evm used in
   let univs = Evd.univ_entry ~poly evm in
@@ -257,9 +233,9 @@ let declare_constant id body ty ~poly ~kind evd =
   let cst = Declare.declare_constant ~name:id (Declare.DefinitionEntry ce) ~kind in
   Flags.if_verbose Feedback.msg_info (str((Id.to_string id) ^ " is defined"));
   if poly then
-    let cstr = EConstr.(mkConstU (cst, EInstance.make (UVars.UContext.instance (Evd.to_universe_context evm)))) in
+    let cstr = EConstr.(mkConstU (cst, EInstance.make (Univ.UContext.instance (Evd.to_universe_context evm)))) in
     cst, (evm0, cstr)
-  else cst, (evm0, EConstr.UnsafeMonomorphic.mkConst cst)
+  else cst, (evm0, EConstr.mkConst cst)
 
 let make_definition ?opaque ?(poly=false) evm ?types b =
   let evm', _, t = make_definition ?opaque ~poly evm ?types b in
@@ -272,7 +248,7 @@ let declare_instance id ~poly evm ctx cl args =
   let term = it_mkLambda_or_LetIn (Option.get c) ctx in
   let typ = EConstr.it_mkProd_or_LetIn t ctx in
   let cst, ecst = declare_constant id term (Some typ) ~poly evm ~kind:Decls.(IsDefinition Instance) in
-  let () = Classes.declare_instance (Global.env ()) evm (Some Hints.empty_hint_info) Hints.SuperGlobal (GlobRef.ConstRef cst) in
+  let () = Classes.Internal.add_instance (fst cl) Hints.empty_hint_info true (GlobRef.ConstRef cst) in
   cst, ecst
 
 let coq_zero = (find_global "nat.zero")
@@ -357,7 +333,7 @@ let get_fresh sigma r = new_global sigma (Lazy.force r)
 
 let get_efresh r evd = e_new_global evd (Lazy.force r)
 
-let is_lglobal env sigma gr c = EConstr.isRefX env sigma (Lazy.force gr) c
+let is_lglobal sigma gr c = EConstr.isRefX sigma (Lazy.force gr) c
 
 open EConstr
 
@@ -434,7 +410,7 @@ let coq_ImpossibleCall evd = find_constant "impossiblecall.class" evd
 
 let unfold_add_pattern =
   lazy (Tactics.unfold_in_concl [(Locus.AllOccurrences,
-			     Evaluable.EvalConstRef (Globnames.destConstRef (Lazy.force coq_add_pattern)))])
+			     Tacred.EvalConstRef (Globnames.destConstRef (Lazy.force coq_add_pattern)))])
 
 let subterm_relation_base = "subterm_relation"
 
@@ -516,19 +492,14 @@ let lift_list l = List.map (Vars.lift 1) l
 (*     | _ -> pars *)
 (*   in aux init n c *)
 
-let is_transparent_constant csts ps c =
-  match Structures.PrimitiveProjections.find_opt c with
-  | None -> Cset.mem c csts
-  | Some p -> PRset.mem p ps
-
-let unfold_head env sigma (ids, csts, ps) c =
+let unfold_head env sigma (ids, csts) c = 
   let rec aux c = 
     match kind sigma c with
     | Var id when Id.Set.mem id ids ->
       (match Environ.named_body id env with
       | Some b -> true, of_constr b
       | None -> false, c)
-    | Const (cst,u) when is_transparent_constant csts ps cst ->
+    | Const (cst,u) when Cset.mem cst csts ->
 	    true, of_constr (Environ.constant_value_in env (cst, EInstance.kind sigma u))
     | App (f, args) ->
       (match aux f with
@@ -557,25 +528,23 @@ open CErrors
 
 let unfold_head env sigma db t =
   let st =
-    List.fold_left (fun (i,c,p) dbname -> 
+    List.fold_left (fun (i,c) dbname -> 
       let db = try Hints.searchtable_map dbname 
         with Not_found -> user_err (str "Unknown database " ++ str dbname)
       in
-      let (ids, csts, ps) = Hints.Hint_db.unfolds db in
-        (Id.Set.union ids i, Cset.union csts c, PRset.union ps p)) (Id.Set.empty, Cset.empty, PRset.empty) db
+      let (ids, csts) = Hints.Hint_db.unfolds db in
+        (Id.Set.union ids i, Cset.union csts c)) (Id.Set.empty, Cset.empty) db
   in
   unfold_head env sigma st t
 
 let autounfold_heads db db' cl =
   Proofview.Goal.enter begin fun gl ->
-  let env = Proofview.Goal.env gl in
-  let sigma = Proofview.Goal.sigma gl in
   let eq = 
     (match cl with Some (id, _) -> pf_get_hyp_typ id gl | None -> pf_concl gl) 
   in
   let did, c' = 
     match kind (project gl) eq with
-    | App (f, [| ty ; x ; y |]) when EConstr.isRefX env sigma (Lazy.force logic_eq_type) f ->
+    | App (f, [| ty ; x ; y |]) when EConstr.isRefX (project gl) (Lazy.force logic_eq_type) f ->
       let did, x' = unfold_head (pf_env gl) (project gl) db x in
       let did', y' = unfold_head (pf_env gl) (project gl) db' y in
       did && did', EConstr.mkApp (f, [| ty ; x' ; y' |])
@@ -585,7 +554,7 @@ let autounfold_heads db db' cl =
       match cl with
       | Some hyp -> change_in_hyp ~check:true None (make_change_arg c') hyp
       | None -> convert_concl ~cast:false ~check:false c' DEFAULTcast
-    else tclFAIL (str "Nothing to unfold")
+    else tclFAIL 0 (str "Nothing to unfold")
   end
 
 type hintdb_name = string
@@ -599,11 +568,23 @@ let dbs_of_constrs = List.map db_of_constr
 
 (** Bindings to Coq *)
 
+let below_tactics_path =
+  DirPath.make (List.map Id.of_string ["Below";"Equations"])
+
+let below_tac s =
+  KerName.make (MPfile below_tactics_path) (Label.make s)
+
 let tacvar_arg h =
   let ipat = Genarg.in_gen (Genarg.rawwit Tacarg.wit_intro_pattern) 
     (CAst.make @@ Tactypes.IntroNaming (Namegen.IntroIdentifier h)) in
     TacGeneric (None, ipat)
 
+let rec_tac h h' = 
+  CAst.(make @@ TacArg(TacCall(
+      make
+        (qualid_of_string "Equations.Below.rec",
+         [tacvar_arg h'; ConstrMayEval (Genredexpr.ConstrTerm h)]))))
+
 let rec_wf_tac h n h' rel =
   CAst.(make @@ TacArg(TacCall(make
     (qualid_of_string "Equations.Subterm.rec_wf_eqns_rel",
@@ -669,7 +650,7 @@ let call_tac_on_ref tac c =
   let tac = Locus.ArgArg (dummy_loc, tac) in
   let val_reference = Geninterp.val_tag (Genarg.topwit Stdarg.wit_constr) in
   (* This is a hack to avoid generating useless universes *)
-  let c = Constr.mkRef (c, UVars.Instance.empty) in
+  let c = Constr.mkRef (c, Univ.Instance.empty) in
   let c = Geninterp.Val.inject val_reference (EConstr.of_constr c) in
   let ist = Geninterp.{ lfun = Names.Id.Map.add var c Names.Id.Map.empty;
                         extra = Geninterp.TacStore.empty; poly = false } in
@@ -798,9 +779,10 @@ let decompose_indapp sigma f args =
   | _ -> f, args
 
 let e_conv env evdref t t' =
-  match Reductionops.infer_conv env !evdref ~pb:Conversion.CONV t t' with
+  match Reductionops.infer_conv env !evdref ~pb:Reduction.CONV t t' with
   | Some sigma -> (evdref := sigma; true)
   | None -> false
+  | exception Reduction.NotConvertible -> false
       
 let deps_of_var sigma id env =
   Environ.fold_named_context
@@ -842,9 +824,6 @@ let move_after_deps id c =
     Tactics.move_hyp id (Logic.MoveAfter first)
   in Proofview.Goal.enter enter
 
-let assert_replacing id c ty =
-  tclTHENFIRST (assert_before_replacing id ty) (exact_no_check c)
-
 let observe s (tac : unit Proofview.tactic) =
   let open Proofview.Notations in
   let open Proofview in
@@ -863,7 +842,8 @@ let observe s (tac : unit Proofview.tactic) =
             (Proofview.numgoals >>= fun gls ->
              if gls = 0 then (Feedback.msg_debug (str s ++ str " succeeded"); Proofview.tclUNIT ())
              else Proofview.Goal.enter begin fun gl ->
-              let () = Feedback.msg_debug (str "Subgoal: " ++ Printer.Debug.pr_goal gl) in
+              let gl = { Evd.it = Proofview.Goal.goal gl; Evd.sigma = Proofview.Goal.sigma gl } in
+              let () = Feedback.msg_debug (str "Subgoal: " ++ Printer.pr_goal gl) in
               Proofview.tclUNIT ()
              end))
          (fun iexn -> Feedback.msg_debug
@@ -918,17 +898,17 @@ let named_of_rel_context ?(keeplets = false) default l =
     List.fold_right
       (fun decl (subst, args, ids, ctx) ->
         let decl = Context.Rel.Declaration.map_constr (substl subst) decl in
-    let id = match get_name decl with Anonymous -> default () | Name id -> id in
-    let d = Named.Declaration.of_rel_decl (fun _ -> id) decl in
+	let id = match get_name decl with Anonymous -> default () | Name id -> id in
+        let d = Named.Declaration.of_tuple (annotR id, get_value decl, get_type decl) in
 	let args = if keeplets ||Context.Rel.Declaration.is_local_assum decl then mkVar id :: args else args in
 	  (mkVar id :: subst, args, id :: ids, d :: ctx))
       l ([], [], [], [])
   in acc, rev args, ctx
 
-let rel_of_named_context sigma ctx = 
+let rel_of_named_context ctx = 
   List.fold_right (fun decl (ctx',subst) ->
       let (n, b, t) = to_named_tuple decl in
-      let decl = make_def (map_annot (fun n -> Name n) n) (Option.map (subst_vars sigma subst) b) (subst_vars sigma subst t) in
+      let decl = make_def (map_annot (fun n -> Name n) n) (Option.map (subst_vars subst) b) (subst_vars subst t) in
       (decl :: ctx', n.binder_name :: subst)) ctx ([],[])
 
 let empty_hint_info = Hints.empty_hint_info
@@ -988,13 +968,13 @@ let map_named_declaration = Context.Named.Declaration.map_constr
 let map_named_context = Context.Named.map
 let lookup_named = Context.Named.lookup
 
-let subst_in_named_ctx sigma (n : Id.t) (c : constr) (ctx : EConstr.named_context) : EConstr.named_context =
+let subst_in_named_ctx (n : Id.t) (c : constr) (ctx : EConstr.named_context) : EConstr.named_context =
   let rec aux after = function
     | [] -> []
     | decl :: before ->
        let name = get_id decl in
        if Id.equal name n then (rev after) @ before
-       else aux (map_named_declaration (replace_vars sigma [n,c]) decl :: after)
+       else aux (map_named_declaration (replace_vars [n,c]) decl :: after)
                 before
   in aux [] ctx
 
@@ -1010,6 +990,12 @@ let print_error e = CErrors.print e
 let nf_betadeltaiota = nf_all
 let anomaly ?label pp = CErrors.anomaly ?label pp
 
+let evar_declare sign ev ty ?src evm =
+  let evi = Evd.make_evar sign ty in
+  let evi = match src with Some src -> { evi with Evd.evar_source = src }
+                         | None -> evi in
+  Evd.add evm ev evi
+
 let new_evar env evm ?src ty =
   Evarutil.new_evar env evm ?src ty
 
@@ -1023,10 +1009,10 @@ let evar_absorb_arguments = Evardefine.evar_absorb_arguments
 let hintdb_set_transparency cst b db =
   let locality = if Global.sections_are_opened () then Hints.Local else Hints.SuperGlobal in
   Hints.add_hints ~locality [db] 
-    (Hints.HintsTransparencyEntry (Hints.HintsReferences [Evaluable.EvalConstRef cst], b))
+    (Hints.HintsTransparencyEntry (Hints.HintsReferences [Tacred.EvalConstRef cst], b))
 
 (* Call the really unsafe is_global test, we use this on evar-open terms too *)
-let is_global = EConstr.isRefX
+let is_global sigma f ec = EConstr.isRefX sigma f ec
 
 let constr_of_global_univ sigma u = of_constr (Constr.mkRef (from_peuniverses sigma u))
 
@@ -1044,7 +1030,7 @@ let decompose_appvect sigma t =
 
 let dest_ind_family fam =
   let ind, fam = Inductiveops.dest_ind_family fam in
-  ind, fam
+  to_peuniverses ind, List.map of_constr fam
 
 (* XXX: EConstr-versions fo these functions really needed XXX *)
 let to_constr = to_constr ~abort_on_undefined_evars:false
@@ -1089,30 +1075,33 @@ let evd_comb1 f evd x =
 (* Universe related functions *)
 
 let nonalgebraic_universe_level_of_universe env sigma u =
-  match ESorts.kind sigma u with
-  | Sorts.Set | Sorts.Prop | Sorts.SProp ->
-    sigma, Univ.Level.set, u
-  | Sorts.Type u0 | Sorts.QSort (_, u0) ->
-    match Univ.Universe.level u0 with
-    | Some l ->
+  match Univ.Universe.level u with
+  | Some l ->
+    if Univ.Level.is_small l then sigma, l, u
+    else
       (match Evd.universe_rigidity sigma l with
       | Evd.UnivFlexible true ->
-        Evd.make_nonalgebraic_variable sigma l, l, ESorts.make @@ Sorts.sort_of_univ @@ Univ.Universe.make l
+        Evd.make_nonalgebraic_variable sigma l, l, Univ.Universe.make l
       | _ -> sigma, l, u)
-    | None ->
-      let sigma, l = Evd.new_univ_level_variable Evd.univ_flexible sigma in
-      let ul = ESorts.make @@ Sorts.sort_of_univ @@ Univ.Universe.make l in
-      let sigma = Evd.set_leq_sort env sigma u ul in
-      sigma, l, ul
+  | _ ->
+    let sigma, l = Evd.new_univ_level_variable Evd.univ_flexible sigma in
+    let ul = Univ.Universe.make l in
+    let sigma = Evd.set_leq_sort env sigma (Sorts.sort_of_univ u) (Sorts.sort_of_univ ul) in
+    sigma, l, ul
 
 let instance_of env sigma ?argu goalu =
   let sigma, goall, goalu = nonalgebraic_universe_level_of_universe env sigma goalu in
+  let sigma, goall, goalu =
+    if Univ.Level.is_prop goall then
+      sigma, Univ.Level.set, Univ.Universe.make Univ.Level.set
+    else sigma, goall, goalu
+  in
   let inst =
     match argu with
     | Some equ ->
       let equ = EConstr.EInstance.kind sigma equ in
-      let quals, equarray = UVars.Instance.to_array equ in
-      EConstr.EInstance.make (UVars.Instance.of_array (quals, Array.append equarray [| goall |]))
-    | None -> EConstr.EInstance.make (UVars.Instance.of_array ([||], [| goall |]))
+      let equarray = Univ.Instance.to_array equ in
+      EConstr.EInstance.make (Univ.Instance.of_array (Array.append equarray [| goall |]))
+    | None -> EConstr.EInstance.make (Univ.Instance.of_array [| goall |])
   in
   sigma, inst, goalu
