@@ -220,7 +220,7 @@ val logic_signature_pack : lazy_ref
 
 val get_fresh : Evd.evar_map -> lazy_ref -> Evd.evar_map * constr
 val get_efresh : lazy_ref -> esigma -> constr
-val is_lglobal : Environ.env -> Evd.evar_map -> lazy_ref -> EConstr.constr -> bool
+val is_lglobal : Evd.evar_map -> lazy_ref -> EConstr.constr -> bool
 
 val coq_sigma : lazy_ref
 val coq_sigmaI : lazy_ref
@@ -289,6 +289,8 @@ val unfold_add_pattern : unit Proofview.tactic lazy_t
 
 val observe : string -> unit Proofview.tactic -> unit Proofview.tactic
   
+val below_tactics_path : Names.DirPath.t
+val below_tac : string -> Names.KerName.t
 val unfold_recursor_tac : unit -> unit Proofview.tactic
 val unfold_recursor_ext_tac : unit -> unit Proofview.tactic
 val equations_tac : unit -> unit Proofview.tactic
@@ -308,6 +310,10 @@ val do_empty_tac : Names.Id.t -> unit Proofview.tactic
 val depelim_nosimpl_tac : Names.Id.t -> unit Proofview.tactic
 val simpl_dep_elim_tac : unit -> unit Proofview.tactic
 val depind_tac : Names.Id.t -> unit Proofview.tactic
+val rec_tac :            Genredexpr.r_trm ->
+                         Names.Id.t ->
+                         Tacexpr.r_dispatch Tacexpr.gen_tactic_expr
+
 val rec_wf_tac :            Genredexpr.r_trm -> Genredexpr.r_trm ->
            Names.Id.t -> Genredexpr.r_trm ->
                          Tacexpr.r_dispatch Tacexpr.gen_tactic_expr
@@ -335,26 +341,25 @@ val ident_of_smart_global :
   Libnames.qualid Constrexpr.or_by_notation -> Id.t
 
 val move_after_deps : Names.Id.t -> constr -> unit Proofview.tactic
-val assert_replacing : Names.Id.t -> constr -> types -> unit Proofview.tactic
 
 val extended_rel_vect : int -> rel_context -> constr array
 val extended_rel_list : int -> rel_context -> constr list
-val to_tuple : rel_declaration -> Names.Name.t binder_annot * constr option * constr
-val to_named_tuple : named_declaration -> Names.Id.t binder_annot * constr option * constr
-val of_tuple : Names.Name.t binder_annot * constr option * constr -> rel_declaration
-val of_named_tuple : Names.Id.t binder_annot * constr option * constr -> named_declaration
+val to_tuple : rel_declaration -> Names.Name.t Context.binder_annot * constr option * constr
+val to_named_tuple : named_declaration -> Names.Id.t Context.binder_annot * constr option * constr
+val of_tuple : Names.Name.t Context.binder_annot * constr option * constr -> rel_declaration
+val of_named_tuple : Names.Id.t Context.binder_annot * constr option * constr -> named_declaration
 
 val get_type : rel_declaration -> constr
 val get_name : rel_declaration -> Names.Name.t
-val get_annot : rel_declaration -> Names.Name.t binder_annot
+val get_annot : rel_declaration -> Names.Name.t Context.binder_annot
 val get_value : rel_declaration -> constr option
-val make_assum : Names.Name.t binder_annot -> constr -> rel_declaration
-val make_def : Names.Name.t binder_annot -> constr option -> constr -> rel_declaration
-val make_named_def : Names.Id.t binder_annot -> constr option -> constr -> named_declaration
-val to_context : (Names.Name.t binder_annot * constr option * constr) list -> rel_context
+val make_assum : Names.Name.t Context.binder_annot -> constr -> rel_declaration
+val make_def : Names.Name.t Context.binder_annot -> constr option -> constr -> rel_declaration
+val make_named_def : Names.Id.t Context.binder_annot -> constr option -> constr -> named_declaration
+val to_context : (Names.Name.t Context.binder_annot * constr option * constr) list -> rel_context
 
 val named_of_rel_context : ?keeplets:bool -> (unit -> Names.Id.t) -> rel_context -> EConstr.t list * constr list * named_context
-val rel_of_named_context : Evd.evar_map -> named_context -> rel_context * Names.Id.t list
+val rel_of_named_context : named_context -> rel_context * Names.Id.t list
 val subst_rel_context : int -> EConstr.t list -> rel_context -> rel_context
 val get_id : named_declaration -> Names.Id.t
 val get_named_type : named_declaration -> constr
@@ -385,7 +390,11 @@ val subst_telescope : constr -> rel_context -> rel_context
 val subst_in_ctx : int -> constr -> rel_context -> rel_context
 val set_in_ctx : int -> constr -> rel_context -> rel_context
 val subst_in_named_ctx :
-  Evd.evar_map -> Names.Id.t -> constr -> named_context -> named_context
+  Names.Id.t -> constr -> named_context -> named_context
+
+val evar_declare : named_context_val ->
+  Evar.t -> 
+  EConstr.types -> ?src:(Evar_kinds.t Loc.located) -> Evd.evar_map -> Evd.evar_map
 
 val new_evar :            Environ.env ->
            Evd.evar_map ->
@@ -395,7 +404,7 @@ val new_evar :            Environ.env ->
 val new_type_evar :            Environ.env ->
            Evd.evar_map -> 
            ?src:Evar_kinds.t Loc.located -> Evd.rigid ->
-           Evd.evar_map * (constr * ESorts.t)
+           Evd.evar_map * (constr * Sorts.t)
 
 val empty_hint_info : 'a Typeclasses.hint_info_gen
 
@@ -409,10 +418,10 @@ val hintdb_set_transparency :
   Constant.t -> bool -> Hints.hint_db_name -> unit
   
 (** To add to the API *)
-val to_peuniverses : 'a UVars.puniverses -> 'a peuniverses
-val from_peuniverses : Evd.evar_map -> 'a peuniverses -> 'a UVars.puniverses
+val to_peuniverses : 'a Univ.puniverses -> 'a peuniverses
+val from_peuniverses : Evd.evar_map -> 'a peuniverses -> 'a Univ.puniverses
 
-val is_global : Environ.env -> Evd.evar_map -> Names.GlobRef.t -> constr -> bool
+val is_global : Evd.evar_map -> Names.GlobRef.t -> constr -> bool
 val constr_of_global_univ : Evd.evar_map -> Names.GlobRef.t peuniverses -> constr
 
 val rel_vect : int -> int -> constr array
@@ -437,13 +446,10 @@ val splay_prod_n_assum : env -> Evd.evar_map -> int -> types -> rel_context * ty
 
 (* Universes *)
 val nonalgebraic_universe_level_of_universe :
-  Environ.env -> Evd.evar_map -> ESorts.t -> Evd.evar_map * Univ.Level.t * ESorts.t
+  Environ.env -> Evd.evar_map -> Univ.Universe.t -> Evd.evar_map * Univ.Level.t * Univ.Universe.t
 val instance_of :
   Environ.env ->
   Evd.evar_map ->
   ?argu:EConstr.EInstance.t ->
-  ESorts.t ->
-  Evd.evar_map * EConstr.EInstance.t * ESorts.t
-
-val collapse_term_qualities : UState.t -> Constr.t -> Constr.t
-(* Hack to prevent sending terms with unbound qualities to the kernel *)
+  Univ.Universe.t ->
+  Evd.evar_map * EConstr.EInstance.t * Univ.Universe.t
