@@ -45,27 +45,27 @@ let find_helper_info env sigma info f =
 	info.helpers_info
   with Not_found -> anomaly (str"Helper not found while proving induction lemma.")
 
-let simp_transparent_state () =
-  Hints.Hint_db.transparent_state (Hints.searchtable_map "simp")
+let below_transparent_state () =
+  Hints.Hint_db.transparent_state (Hints.searchtable_map "Below")
 
 let simpl_star =
   tclTHEN simpl_in_concl (onAllHyps (fun id -> simpl_in_hyp (id, Locus.InHyp)))
 
-let eauto_with_rec ?depth ?(strategy=Class_tactics.Dfs) l =
-  Class_tactics.typeclasses_eauto ~depth ~st:(simp_transparent_state ()) 
-    ~strategy (l@["subterm_relation"; "rec_decision"])
+let eauto_with_below ?depth ?(strategy=Class_tactics.Dfs) l =
+  Class_tactics.typeclasses_eauto ~depth ~st:(below_transparent_state ()) 
+    ~strategy (l@["subterm_relation"; "Below"; "rec_decision"])
     
 let wf_obligations_base info =
   info.base_id ^ "_wf_obligations"
 
 let simp_eqns l =
   tclREPEAT
-    (tclTHENLIST [Autorewrite.autorewrite tclIDTAC l; tclTRY (eauto_with_rec ("simp" :: l))])
+    (tclTHENLIST [Autorewrite.autorewrite tclIDTAC l; tclTRY (eauto_with_below l)])
 
 let simp_eqns_in clause l =
   tclREPEAT (tclTHENLIST 
 		[Autorewrite.auto_multi_rewrite l clause;
-		 tclTRY (eauto_with_rec ("simp" :: l))])
+		 tclTRY (eauto_with_below l)])
 
 let autorewrites b = 
   tclREPEAT (Autorewrite.autorewrite tclIDTAC [b])
@@ -81,7 +81,7 @@ let _rewrite_try_change tac =
           let env = Proofview.Goal.env gl in
           let sigma = Proofview.Goal.sigma gl in
           let concl' = Proofview.Goal.concl gl in
-          match Reductionops.infer_conv ~pb:Conversion.CONV env sigma concl concl' with
+          match Reductionops.infer_conv ~pb:Reduction.CONV env sigma concl concl' with
           | Some _ -> Proofview.tclZERO (RewriteSucceeded concl')
           | None -> Proofview.tclUNIT ())))
     (function
@@ -92,13 +92,13 @@ let autorewrite_one b =
   let rew_rules = Autorewrite.find_rewrites b in
   let rec aux rules =
     match rules with
-    | [] -> tclFAIL (str"Couldn't rewrite")
+    | [] -> tclFAIL 0 (str"Couldn't rewrite")
     | r :: rules ->
-       let global, _univs = Constr.destRef (snd @@ Autorewrite.RewRule.rew_lemma r) in
+       let global, _univs = Constr.destRef r.Autorewrite.rew_lemma in
        let tac =
          Proofview.tclBIND
          (pf_constr_of_global global)
-         (if (Autorewrite.RewRule.rew_l2r r) then Equality.rewriteLR else Equality.rewriteRL)
+         (if r.Autorewrite.rew_l2r then Equality.rewriteLR else Equality.rewriteRL)
        in
        Proofview.tclOR tac
          (fun e -> if !debug then Feedback.msg_debug (str"failed"); aux rules)
@@ -107,7 +107,7 @@ let autorewrite_one b =
 let revert_last =
   Proofview.Goal.enter (fun gl ->
       let hyp = pf_last_hyp gl in
-      Generalize.revert [get_id hyp])
+      revert [get_id hyp])
 
 (** fix generalization *)
 
@@ -137,13 +137,12 @@ let mutual_fix li l =
   let open Proofview in
   let mfix env sigma gls =
     let gls = List.map Proofview.drop_state gls in
-    let infos = List.map (fun ev -> Evd.find_undefined sigma ev) gls in
-    let types = List.map (fun evi -> Evd.evar_relevance evi, Evd.evar_concl evi) infos in
+    let types = List.map (fun ev -> Evd.evar_concl (Evd.find sigma ev)) gls in
     let env =
-      let ctxs = List.map (fun evi -> EConstr.Unsafe.to_named_context @@
-                            Evd.evar_context evi) infos in
+      let ctxs = List.map (fun ev -> EConstr.Unsafe.to_named_context @@
+                            Evd.evar_context (Evd.find sigma ev)) gls in
       let fst, rest = List.sep_last ctxs in
-      if List.for_all (fun y -> Context.Named.equal Sorts.relevance_equal Constr.equal fst y) rest then
+      if List.for_all (fun y -> Context.Named.equal Constr.equal fst y) rest then
         Environ.push_named_context fst env
       else env
     in
@@ -168,18 +167,18 @@ let mutual_fix li l =
                             int lengoals ++ str(String.plural lengoals " subgoal"))
     in
     let all = CList.map3 (fun id n ar -> (id,n,ar)) li l types in
-    let (_, n, (_, ar)) = List.hd all in
+    let (_, n, ar) = List.hd all in
     let (sp, u) = check_mutind env sigma n ar in
     let rec mk_sign sign = function
       | [] -> sign
-      | (f, n, (r, ar)) :: oth ->
+      | (f, n, ar) :: oth ->
          let (sp', u')  = check_mutind env sigma n ar in
          if not (Environ.QMutInd.equal env sp sp') then
            error "Fixpoints should be on the same mutual inductive declaration.";
          if try ignore (Context.Named.lookup f sign); true with Not_found -> false then
            CErrors.user_err
                     (str "Name " ++ pr_id f ++ str " already used in the environment");
-         mk_sign (LocalAssum (make_annot f (ERelevance.kind sigma r), EConstr.to_constr sigma ar) :: sign) oth
+         mk_sign (LocalAssum (annotR f, EConstr.to_constr sigma ar) :: sign) oth
     in
     let sign = mk_sign (Environ.named_context env) all in
     let idx = Array.map_of_list pred l in
@@ -189,9 +188,8 @@ let mutual_fix li l =
       Refine.refine ~typecheck:false
       (fun sigma ->
         let nenv = Environ.reset_with_named_context (Environ.val_of_named_context sign) env in
-        let types = List.map snd types in
         let (sigma, evs) = mk_holes nenv sigma types in
-        let evs = Array.map_of_list (Vars.subst_vars sigma (List.rev li)) evs in
+        let evs = Array.map_of_list (Vars.subst_vars (List.rev li)) evs in
         let types = Array.of_list types in
         let decl = (nas,types,evs) in
         let () = body := (fun i -> mkFix ((idx,i),decl)) in
@@ -210,8 +208,8 @@ let mutual_fix li l =
 let check_guard gls env sigma =
   let gl = Proofview.drop_state (List.hd gls) in
   try
-    let EvarInfo evi = Evd.find sigma gl in
-    match Evd.evar_body evi with
+    let evi = Evd.find sigma gl in
+    match evi.Evd.evar_body with
     | Evd.Evar_defined b -> Inductiveops.control_only_guard (Evd.evar_env env evi) sigma b; true
     | Evd.Evar_empty -> true
   with Type_errors.TypeError _ -> false
@@ -222,7 +220,7 @@ let find_helper_arg env sigma info f args =
       
 let find_splitting_var sigma pats var constrs =
   let rec find_pat_var p c =
-    match p, decompose_app_list sigma c with
+    match p, decompose_app sigma c with
     | PRel i, (c, l) when i = var -> Some (destVar sigma c)
     | PCstr (c, ps), (f,l) -> aux ps l
     | _, _ -> None
@@ -263,8 +261,8 @@ let map_opt_split f s =
   | Some s -> f s
 
 let solve_ind_rec_tac info =
-  observe "solve_ind_rec_tac"
-    (eauto_with_rec ~depth:20 ~strategy:Class_tactics.Bfs [info.base_id; wf_obligations_base info])
+  observe "eauto with below"
+    (eauto_with_below ~depth:20 ~strategy:Class_tactics.Bfs [info.base_id; wf_obligations_base info])
 
 let change_in_app f args idx arg =
   let args' = Array.copy args in
@@ -277,9 +275,9 @@ let hyps_after sigma env pos args =
 
 let simpl_of csts =
   let opacify () = List.iter (fun (cst,_) ->
-    Global.set_strategy (Conv_oracle.EvalConstRef cst) Conv_oracle.Opaque) csts
+    Global.set_strategy (ConstKey cst) Conv_oracle.Opaque) csts
   and transp () = List.iter (fun (cst, level) ->
-    Global.set_strategy (Conv_oracle.EvalConstRef cst) level) csts
+    Global.set_strategy (ConstKey cst) level) csts
   in opacify, transp
 
 let gather_subst env sigma ty args len =
@@ -289,7 +287,7 @@ let gather_subst env sigma ty args len =
       | Prod (_, _, ty), a :: args -> a :: aux (subst1 a ty) args (pred n)
       | LetIn (_, b, _, ty), args -> b :: aux (subst1 b ty) args (pred n)
       | _ -> assert false
-  in aux ty (Array.to_list args) len
+  in aux ty args len
 
 let annot_of_rec r = match r.struct_rec_arg with
   | MutualOn (Some (i, _)) -> Some (i + 1)
@@ -333,7 +331,6 @@ let aux_ind_fun info chop nested unfp unfids p =
           let program_tac =
             tclTHEN fixtac (aux chop None [] prog.program_splitting)
           in
-          let ty = EConstr.of_constr @@ collapse_term_qualities (Evd.evar_universe_context sigma) (EConstr.to_constr sigma ty) in
           tclTHEN (assert_by (Name (program_id prog)) ty program_tac)
             (observe "solving nested premises of compute rule"
               (solve_ind_rec_tac info.term_info))
@@ -377,10 +374,10 @@ let aux_ind_fun info chop nested unfp unfids p =
         let sigma = ref (Proofview.Goal.sigma gl) in
         match t with
         | { rec_node = WfRec r } ->
-          let inctx, concl = decompose_prod_n_decls !sigma t.rec_args (concl gl) in
+          let inctx, concl = decompose_prod_n_assum !sigma t.rec_args (concl gl) in
           Refine.refine ~typecheck:false (fun sigma ->
               let evd = ref sigma in
-              let sort = Retyping.get_sort_of env sigma concl in
+              let sort = Sorts.univ_of_sort (Retyping.get_sort_of env sigma concl) in
               let hd, args = decompose_app sigma concl in
               let subst =
                 gather_subst env sigma (Retyping.get_type_of env sigma hd) args (List.length ctx)
@@ -403,7 +400,7 @@ let aux_ind_fun info chop nested unfp unfids p =
                 arity, arg, r.wf_rec_rel
               in
               let _functional_type, functional_type, fix =
-                Covering.wf_fix_constr env evd inctx concl (ESorts.kind !evd sort) arity arg rel
+                Covering.wf_fix_constr env evd inctx concl sort arity arg rel
               in
               (* TODO solve WellFounded evar *)
               let sigma, evar = new_evar env !evd functional_type in
@@ -438,7 +435,7 @@ let aux_ind_fun info chop nested unfp unfids p =
                         (tclTHEN intros (aux chop unfs unfids cs)))]
 
   and aux chop unfs unfids = function
-    | Split (lhs, var, _, splits) ->
+    | Split ((ctx,pats,_ as lhs), var, _, splits) ->
       let splits = List.map_filter (fun x -> x) (Array.to_list splits) in
       let unfs_splits =
         let unfs = map_opt_split destSplit unfs in
@@ -463,14 +460,14 @@ let aux_ind_fun info chop nested unfp unfids p =
                 in
                 let id = find_splitting_var (project gl) pats var pats' in
                 Depelim.dependent_elim_tac (None, id)
-              | _ -> tclFAIL (str"Unexpected goal in functional induction proof")))
+              | _ -> tclFAIL 0 (str"Unexpected goal in functional induction proof")))
            (fun i ->
             Proofview.Goal.enter (fun gl ->
               let split = nth splits (pred i) in
               let unfsplit = Option.map (fun s -> nth s (pred i)) unfs_splits in
               aux chop unfsplit unfids split))))
 
-    | Refined (lhs, refinfo, s) ->
+    | Refined ((ctx, _, _), refinfo, s) ->
       let unfs = map_opt_split destRefined unfs in
       let id = pi1 refinfo.refined_obj in
       let elimtac gl =
@@ -493,12 +490,12 @@ let aux_ind_fun info chop nested unfp unfids p =
             mkApp (indapp, [| fnapp |])
           in
           tclTHENLIST
-            [observe "letin" (letin_pat_tac true None (Name id) (Some sigma, elim) occs);
+            [observe "letin" (letin_pat_tac true None (Name id) (sigma, elim) occs);
              observe "convert concl" (convert_concl ~cast:false ~check:false newconcl DEFAULTcast);
              observe "clear body" (clear_body [id]);
              aux chop unfs unfids s]
         | _ ->
-          tclFAIL (str"Unexpected refinement goal in functional induction proof")
+          tclFAIL 0 (str"Unexpected refinement goal in functional induction proof")
       in
       (observe "refine"
               (tclTHENLIST [ intros;
@@ -507,13 +504,12 @@ let aux_ind_fun info chop nested unfp unfids p =
                                (tclSOLVE [Proofview.Goal.enter elimtac]);
                              (solve_ind_rec_tac info.term_info)]))
 
-    | Compute (lhs, wheres, _, c) ->
-      let lctx = lhs.src_ctx in
+    | Compute ((lctx,_,_), wheres, _, c) ->
       let unfctx, unfswheres =
         let unfs = map_opt_split destWheres unfs in
         match unfs with
         | None -> [], List.map (fun _ -> None) wheres
-        | Some (unfctx, wheres) -> unfctx.src_ctx, List.map (fun w -> Some w) wheres
+        | Some (unfctx, wheres) -> pi1 unfctx, List.map (fun w -> Some w) wheres
       in
       let wheretac =
         if not (List.is_empty wheres) then
@@ -583,14 +579,12 @@ let aux_ind_fun info chop nested unfp unfids p =
                   pr_context_map env Evd.empty (id_subst ctx)));
             let ind = Nametab.locate (qualid_of_ident wherepath) in
             let ty ind =
-              let hd, args = decompose_app_list Evd.empty where_term in
+              let hd, args = decompose_app Evd.empty where_term in
               let indargs = List.filter (fun x -> isVar Evd.empty x) args in
               let rels = extended_rel_list 0 ctx in
               let indargs = List.append indargs rels in
               let app = applistc ind (List.append indargs [applistc where_term rels]) in
-              let ty = it_mkProd_or_LetIn app ctx in
-              let ty = EConstr.of_constr @@ collapse_term_qualities UState.empty (EConstr.Unsafe.to_constr ty) in
-              ty
+              it_mkProd_or_LetIn app ctx
             in
             let tac =
               tclTHEN acc
@@ -611,13 +605,13 @@ let aux_ind_fun info chop nested unfp unfids p =
                 let sigma = Proofview.Goal.sigma gl in
                 let subst =
                   let concl = Proofview.Goal.concl gl in
-                  let hd, args = decompose_app_list sigma concl in
+                  let hd, args = decompose_app sigma concl in
                   let args = drop_last args in
                   let rec collect_vars acc c =
                     let hd, args = decompose_app sigma c in
                     match kind sigma hd with
                     | Var id -> if not (List.mem id acc) then id :: acc else acc
-                    | Construct _ -> Array.fold_left collect_vars acc args
+                    | Construct _ -> List.fold_left collect_vars acc args
                     | _ -> acc
                   in
                   let args_vars = List.fold_left collect_vars [] args in
@@ -661,7 +655,7 @@ let aux_ind_fun info chop nested unfp unfids p =
 
 let pr_subgoals sigma goals =
   let open Pp in
-  let pr g = str "[" ++ Printer.Debug.pr_goal g ++ str "]" in
+  let pr g = str "[" ++ Printer.pr_goal { Evd.it = g ; Evd.sigma = sigma } ++ str "]" in
   str "[" ++ prlist_with_sep fnl pr goals ++ str "]"
 
 let observe_tac s tac =
@@ -670,8 +664,8 @@ let observe_tac s tac =
   else
     tclENV >>= fun env ->
     tclEVARMAP >>= fun sigma ->
-    Proofview.Goal.goals >>= fun gls ->
-    Proofview.Monad.List.map (fun gl -> gl) gls >>= fun gls ->
+    Unsafe.tclGETGOALS >>= fun gls ->
+    let gls = List.map Proofview.drop_state gls in
     Feedback.msg_debug (str"Applying " ++ str s ++ str " on " ++
                           pr_subgoals sigma gls);
     Proofview.tclORELSE
@@ -680,7 +674,8 @@ let observe_tac s tac =
                           if gls = 0 then (Feedback.msg_debug (str s ++ str "succeeded");
                                            Proofview.tclUNIT ())
              else Proofview.Goal.enter begin fun gl ->
-              let () = Feedback.msg_debug (str "Subgoal: " ++ Printer.Debug.pr_goal gl) in
+              let gl = { Evd.it = Proofview.Goal.goal gl; sigma = Proofview.Goal.sigma gl } in 
+              let () = Feedback.msg_debug (str "Subgoal: " ++ Printer.pr_goal gl) in
               Proofview.tclUNIT ()
              end))
       (fun iexn -> Feedback.msg_debug
@@ -902,7 +897,6 @@ let wrap tac before after =
 
 let solve_rec_eq simpltac subst =
   Proofview.Goal.enter begin fun gl ->
-  let env = Proofview.Goal.env gl in
   match kind (project gl) (pf_concl gl) with
   | App (eq, [| ty; x; y |]) ->
       let sigma = project gl in
@@ -910,127 +904,18 @@ let solve_rec_eq simpltac subst =
       (try
         let f_cst, funf_cst =
           List.find (fun (f_cst, funf_cst) ->
-              is_global env sigma (GlobRef.ConstRef f_cst) xf && is_global env sigma (GlobRef.ConstRef funf_cst) yf) subst
+              is_global sigma (GlobRef.ConstRef f_cst) xf && is_global sigma (GlobRef.ConstRef funf_cst) yf) subst
         in
         let unfolds = unfold_in_concl
-            [((Locus.OnlyOccurrences [1]), Evaluable.EvalConstRef f_cst); 
-              ((Locus.OnlyOccurrences [1]), Evaluable.EvalConstRef funf_cst)]
+            [((Locus.OnlyOccurrences [1]), Tacred.EvalConstRef f_cst); 
+              ((Locus.OnlyOccurrences [1]), Tacred.EvalConstRef funf_cst)]
         in tclTHENLIST [unfolds; simpltac; pi_tac ()]
       with Not_found -> tclORELSE reflexivity (congruence_tac 10 []))
   | _ -> reflexivity
   end
 
-type unfold_subst = (Constant.t * Constant.t) list * Splitting.program * Splitting.program
-
-type unfold_trace =
-| UnfSplit of unfold_trace list
-| UnfRefined of refined_node * unfold_trace
-| UnfComputeProgram of (Splitting.program * Splitting.program * EConstr.t * Id.t) list * EConstr.rel_context
-| UnfComputeEmpty of Id.t
-
-type reckind =
-| RecWfPlain of unfold_trace
-| RecWfWithFunext
-| RecStruct of (int * int) option
-
-let compute_unfold_trace env sigma where_map split unfold_split =
-  let rec aux split unfold_split =
-    match split, unfold_split with
-    | Split (_, _, _, splits), Split (lhs, var, _, unfsplits) ->
-      let ctx = lhs.src_ctx in
-      if is_primitive env sigma ctx (pred var) then
-        aux (Option.get (Array.hd splits)) (Option.get (Array.hd unfsplits))
-      else
-        let splits = List.map_filter (fun x -> x) (Array.to_list splits) in
-        let unfsplits = List.map_filter (fun x -> x) (Array.to_list unfsplits) in
-        let trace = List.map2 (fun split unfsplit -> aux split unfsplit) splits unfsplits in
-        UnfSplit trace
-    | _, Mapping (lhs, s) -> aux split s
-    | Refined (_, _, s), Refined (lhs, refinfo, unfs) ->
-      UnfRefined (refinfo, aux s unfs)
-    | Compute (_, wheres, _, RProgram _), Compute (lhs, unfwheres, _, RProgram _) ->
-      let () = assert (List.length wheres = List.length unfwheres) in
-      let map w unfw =
-        let assoc, id, _ =
-          try PathMap.find unfw.where_path where_map
-          with Not_found -> assert false
-        in
-        let wp = w.where_program in
-        let unfwp = unfw.where_program in
-        (wp, unfwp, assoc, id)
-      in
-      let data = List.map2 map wheres unfwheres in
-      UnfComputeProgram (data, lhs.src_ctx)
-    | Compute (_, _, _, _), Compute (lhs, _, _, REmpty (id, sp)) ->
-      let d = nth lhs.src_ctx (pred id) in
-      let id = Name.get_id (get_name d) in
-      UnfComputeEmpty id
-    | _, _ -> assert false
-  in
-  aux split unfold_split
-
-let get_program_reckind env sigma where_map p = match p.program_rec with
-| None -> None
-| Some r ->
-  let k = match r with
-  | { rec_node = WfRec _ } ->
-    if !Equations_common.equations_with_funext then
-      RecWfWithFunext
-    else
-      let trace = compute_unfold_trace env sigma where_map p.program_splitting p.program_splitting in
-      RecWfPlain trace
-  | { rec_node = StructRec sr } ->
-    match annot_of_rec sr with
-    | Some annot -> RecStruct (Some (r.rec_args, annot))
-    | None -> RecStruct None
-  in
-  Some k
-
-let extract_subprogram_trace env sigma where_map trace =
-  let rec aux subst accu trace = match trace with
-  | UnfSplit traces ->
-    List.fold_left (fun accu trace -> aux subst accu trace) accu traces
-  | UnfRefined (_, trace) -> aux subst accu trace
-  | UnfComputeEmpty _ -> accu
-  | UnfComputeProgram (data, lctx) ->
-    let fold accu (wp, unfwp, assoc, id) =
-      let kind = get_program_reckind env sigma where_map wp in
-      let nsubst, etrace = match kind with
-      | None -> subst, None
-      | Some kind ->
-        let f_cst = headcst sigma wp.program_term in
-        let funf_cst = headcst sigma unfwp.program_term in
-        let etrace = match kind with
-        | RecWfPlain etrace -> Some etrace
-        | RecWfWithFunext | RecStruct _ -> None
-        in
-        (f_cst, funf_cst) :: subst, etrace
-      in
-      let ntrace = compute_unfold_trace env sigma where_map wp.program_splitting unfwp.program_splitting in
-      let accu = match etrace with None -> accu | Some etrace -> aux nsubst accu etrace in
-      let accu = aux nsubst accu ntrace in
-      let evd = ref sigma in
-      let ty =
-        let ctx = unfwp.program_info.program_sign in
-        let len = List.length ctx - List.length lctx in
-        let newctx, oldctx = List.chop len ctx in
-        let lhs = mkApp (lift len assoc, extended_rel_vect 0 newctx) in
-        let rhs = mkApp (unfwp.program_term, extended_rel_vect 0 ctx) in
-        let eq = mkEq env evd unfwp.program_info.program_arity lhs rhs in
-        it_mkProd_or_LetIn eq ctx
-      in
-      let uctx = Evd.evar_universe_context !evd in
-      (id, (subst, wp, unfwp), uctx, ty) :: accu
-    in
-    List.fold_left fold accu data
-  in
-  List.rev (aux [] [] trace)
-
-let extract_subprograms env sigma where_map p unfp =
-  let trace = compute_unfold_trace env sigma where_map p.program_splitting unfp.program_splitting in
-  extract_subprogram_trace env sigma where_map trace
-
-let prove_unfolding info where_map f_cst funf_cst subst base unfold_base trace =
+let prove_unfolding_lemma info where_map f_cst funf_cst p unfp =
+  Proofview.Goal.enter begin fun gl ->
   let depelim h = Depelim.dependent_elim_tac (None, h) (* depelim_tac h *) in
   let helpercsts = List.map (fun (cst, i) -> cst) info.helpers_info in
   let opacify, transp = simpl_of ((destConstRef (Lazy.force coq_hidebody), Conv_oracle.transparent)
@@ -1038,18 +923,74 @@ let prove_unfolding info where_map f_cst funf_cst subst base unfold_base trace =
   let opacified tac = wrap tac opacify transp in
   let transparent tac = wrap tac transp opacify in
   let simpltac = opacified (simpl_equations_tac ()) in
+  let my_simpl = opacified simpl_in_concl in
   let unfolds base base' =
     tclTHEN (autounfold_heads [base] [base'] None)
-    (Tactics.reduct_in_concl ~cast:false ~check:false ((Reductionops.clos_norm_flags RedFlags.betazeta), DEFAULTcast))
+    (Tactics.reduct_in_concl ~cast:false ~check:false ((Reductionops.clos_norm_flags CClosure.betazeta), DEFAULTcast))
   in
   let solve_rec_eq subst = solve_rec_eq simpltac subst in
   let solve_eq subst = observe "solve_eq" (tclORELSE (transparent reflexivity) (solve_rec_eq subst)) in
   let abstract tac = (* Abstract.tclABSTRACT None *) tac in
+  let rec aux_program subst p unfp =
+    Proofview.Goal.enter (fun gl ->
+        let sigma = Proofview.Goal.sigma gl in
+        let f_cst = headcst sigma p.program_term
+        and funf_cst = headcst sigma unfp.program_term in
+        let unfolds =
+          tclTHENLIST
+            [unfold_in_concl
+                     [Locus.OnlyOccurrences [1], Tacred.EvalConstRef f_cst;
+                      (Locus.OnlyOccurrences [1], Tacred.EvalConstRef funf_cst)];
+             my_simpl]
+        in
+        let set_opaque () =
+          Global.set_strategy (ConstKey f_cst) Conv_oracle.Opaque;
+          Global.set_strategy (ConstKey funf_cst) Conv_oracle.Opaque;
+        in
+        let subst, fixtac, extgl =
+          match p.program_rec with
+          | Some r ->
+            let fixtac, extgl = match r with
+              | { rec_node = WfRec _ } ->
+                if !Equations_common.equations_with_funext then
+                  tclTHENLIST [unfolds; unfold_recursor_ext_tac ()], false
+                else
+                  tclTHENLIST [unfolds; unfold_recursor_tac ()], true
+              | { rec_node = StructRec sr } ->
+                match annot_of_rec sr with
+                | Some annot ->
+                  tclTHENLIST [tclDO r.rec_args revert_last;
+                               observe "mutfix" (mutual_fix [] [annot]);
+                               tclDO r.rec_args intro; unfolds], false
+                | None -> Proofview.tclUNIT (), false
+            in ((f_cst, funf_cst) :: subst), fixtac, extgl
+          | _ -> subst, unfolds, false
+        in
+        tclTHENLIST
+          [observe "program before unfold"  intros;
+           if extgl then
+            (tclTHENFIRST
+              (observe "program fixpoint" fixtac)
+              (tclORELSE
+                (tclSOLVE
+                  [Proofview.Goal.enter (fun gl -> set_opaque ();
+                    observe "extensionality proof"
+                    ((aux subst info.base_id info.base_id p.program_splitting p.program_splitting)))])
+                (tclFAIL 0
+                  (Pp.str "Could not prove extensionality automatically"))))
+            else observe "program fixpoint" fixtac;
+            (Proofview.Goal.enter (fun gl -> set_opaque ();
+              (observe "program"
+                ((aux subst info.base_id (info.base_id ^ "_unfold") p.program_splitting unfp.program_splitting)))))])
 
-  let rec aux trace = match trace with
-  | UnfSplit traces ->
+  and aux subst base unfold_base split unfold_split =
+    match split, unfold_split with
+    | Split (_, _, _, splits), Split ((ctx,pats,_), var, _, unfsplits) ->
       observe "split"
         (Proofview.Goal.enter (fun gl ->
+        if is_primitive (pf_env gl) (project gl) ctx (pred var) then
+          aux subst base unfold_base (Option.get (Array.hd splits)) (Option.get (Array.hd unfsplits))
+        else
           match kind (project gl) (pf_concl gl) with
           | App (eq, [| ty; x; y |]) ->
             let sigma = project gl in
@@ -1059,10 +1000,18 @@ let prove_unfolding info where_map f_cst funf_cst subst base unfold_base trace =
               c, tclIDTAC
             in
             let id = destVar sigma (fst (decompose_app sigma c)) in
-            let k i = nth traces (pred i) in
-            abstract (local_tclTHEN_i (depelim id) (fun i -> (tclTHENLIST [unfolds; simpltac; aux (k i)])))
-          | _ -> tclFAIL (str"Unexpected unfolding goal")))
-  | UnfRefined (refinfo, trace) ->
+            let splits = List.map_filter (fun x -> x) (Array.to_list splits) in
+            let unfsplits = List.map_filter (fun x -> x) (Array.to_list unfsplits) in
+            abstract (local_tclTHEN_i (depelim id)
+                                      (fun i -> let split = nth splits (pred i) in
+                        let unfsplit = nth unfsplits (pred i) in
+                        tclTHENLIST [unfolds; simpltac;
+                          aux subst base unfold_base split unfsplit]))
+          | _ -> tclFAIL 0 (str"Unexpected unfolding goal")))
+
+    | _, Mapping (lhs, s) -> aux subst base unfold_base split s
+       
+    | Refined (_, _, s), Refined ((ctx, _, _), refinfo, unfs) ->
         let id = pi1 refinfo.refined_obj in
         let rec reftac () =
           Proofview.Goal.enter begin fun gl ->
@@ -1081,130 +1030,76 @@ let prove_unfolding info where_map f_cst funf_cst subst base unfold_base trace =
                   (letin_tac None (Name id) a2 None Locusops.allHypsAndConcl);
                 clear_body [id]; 
                 observe "unfoldings" (unfolds base unfold_base); 
-                aux trace]
+                aux subst base unfold_base s unfs]
              else tclTHENLIST [unfolds base unfold_base; simpltac; reftac ()]
-          | _ -> tclFAIL (str"Unexpected unfolding lemma goal")
+          | _ -> tclFAIL 0 (str"Unexpected unfolding lemma goal")
           end
         in
       let reftac = observe "refined" (reftac ()) in
       abstract (tclTHENLIST [intros; simpltac; reftac])
-  | UnfComputeProgram (data, lctx) ->
-       let wheretac acc (wp, unfwp, assoc, id) = match Nametab.locate_constant (Libnames.qualid_of_ident id) with
-       | cst ->
+
+    | Compute (_, wheres, _, RProgram _), Compute ((lctx, _, _), unfwheres, _, RProgram c) ->
+       let wheretac acc w unfw =
+         let assoc, id, _ =
+           try PathMap.find unfw.where_path where_map
+           with Not_found -> assert false
+         in
         Proofview.Goal.enter (fun gl ->
          let env = pf_env gl in
          let evd = ref (project gl) in
+         let wp = w.where_program in
+         let unfwp = unfw.where_program in
          (* let () = Feedback.msg_debug (str"Unfold where assoc: " ++
           *                              Printer.pr_econstr_env env !evd assoc) in
           * let () = Feedback.msg_debug (str"Unfold where problem: " ++
           *                              pr_context_map env !evd wp.program_prob) in
           * let () = Feedback.msg_debug (str"Unfold where problem: " ++
           *                              pr_context_map env !evd unfwp.program_prob) in *)
+         let ty =
+           let ctx = unfwp.program_info.program_sign in
+           let len = List.length ctx - List.length lctx in
+           let newctx, oldctx = List.chop len ctx in
+           let lhs = mkApp (lift len assoc, extended_rel_vect 0 newctx) in
+           let rhs = mkApp (unfwp.program_term, extended_rel_vect 0 ctx) in
+           let eq = mkEq env evd unfwp.program_info.program_arity lhs rhs in
+           it_mkProd_or_LetIn eq ctx
+         in
          (* let _ = Typing.type_of env !evd ty in *)
-         let cst = Equations_common.evd_comb1 (Evd.fresh_constant_instance env) evd cst in
-         let cst = EConstr.of_constr (Constr.mkConstU cst) in
-         let ty = Retyping.get_type_of env !evd cst in
+         let progtac = aux_program subst wp unfwp in
          let tac =
            assert_by (Name id) ty
-             (tclTHEN (keep []) (Tactics.exact_check cst))
+             (tclTHEN (keep [])
+                      (Abstract.tclABSTRACT (Some id) progtac))
          in
          tclTHENLIST [Proofview.Unsafe.tclEVARS !evd; tac;
                       Equality.rewriteLR (mkVar id);
                       acc])
-       | exception Not_found -> tclFAIL (Pp.str "Missing subproof " ++ Id.print id)
        in
-       let wheretacs = List.fold_left wheretac tclIDTAC data in
+       let wheretacs =
+         assert(List.length wheres = List.length unfwheres);
+         List.fold_left2 wheretac tclIDTAC wheres unfwheres
+       in
        observe "compute"
          (tclTHENLIST [intros; wheretacs;
                        observe "compute rhs" (tclTRY (unfolds base unfold_base));
                        simpltac; solve_eq subst])
-  | UnfComputeEmpty id ->
-    abstract (depelim id)
+
+    | Compute (_, _, _, _), Compute ((ctx,_,_), _, _, REmpty (id, sp)) ->
+      let d = nth ctx (pred id) in
+      let id = Name.get_id (get_name d) in
+      abstract (depelim id)
+
+    | _, _ -> assert false
   in
-  aux trace
-
-let prove_unfolding_lemma_aux info where_map my_simpl subst p unfp =
-  Proofview.Goal.enter begin fun gl ->
-    let env = Proofview.Goal.env gl in
-    let sigma = Proofview.Goal.sigma gl in
-    let f_cst = headcst sigma p.program_term
-    and funf_cst = headcst sigma unfp.program_term in
-    let unfolds =
-      tclTHENLIST
-        [unfold_in_concl
-                  [Locus.OnlyOccurrences [1], Evaluable.EvalConstRef f_cst;
-                  (Locus.OnlyOccurrences [1], Evaluable.EvalConstRef funf_cst)];
-          my_simpl]
-    in
-    let set_opaque () =
-      Global.set_strategy (Conv_oracle.EvalConstRef f_cst) Conv_oracle.Opaque;
-      Global.set_strategy (Conv_oracle.EvalConstRef funf_cst) Conv_oracle.Opaque;
-    in
-    let kind = get_program_reckind env sigma where_map p in
-    let subst, fixtac, extgl = match kind with
-    | None -> subst, unfolds, None
-    | Some kind ->
-      let fixtac, extgl = match kind with
-        | RecWfPlain etrace ->
-          tclTHENLIST [unfolds; unfold_recursor_tac ()], Some etrace
-        | RecWfWithFunext ->
-          tclTHENLIST [unfolds; unfold_recursor_ext_tac ()], None
-        | RecStruct annot ->
-          match annot with
-          | Some (rec_args, annot) ->
-            tclTHENLIST [tclDO rec_args revert_last;
-                          observe "mutfix" (mutual_fix [] [annot]);
-                          tclDO rec_args intro; unfolds], None
-          | None -> Proofview.tclUNIT (), None
-      in
-      ((f_cst, funf_cst) :: subst), fixtac, extgl
-    in
-(*         let self wp unfwp = aux_program subst wp unfwp in *)
-    let trace = compute_unfold_trace env sigma where_map p.program_splitting unfp.program_splitting in
-    tclTHENLIST
-      [observe "program before unfold"  intros;
-        begin match extgl with
-        | Some etrace ->
-        (tclTHENFIRST
-          (observe "program fixpoint" fixtac)
-          (tclORELSE
-            (tclSOLVE
-              [Proofview.Goal.enter (fun gl -> set_opaque ();
-                observe "extensionality proof"
-                ((prove_unfolding info where_map f_cst funf_cst subst info.base_id info.base_id etrace)))])
-            (tclFAIL
-              (Pp.str "Could not prove extensionality automatically"))))
-        | None -> observe "program fixpoint" fixtac
-        end;
-        (Proofview.Goal.enter (fun gl -> set_opaque ();
-          (observe "program"
-            ((prove_unfolding info where_map f_cst funf_cst subst info.base_id (info.base_id ^ "_unfold") trace)))))]
-  end
-
-let prove_unfolding_lemma info where_map f_cst funf_cst p unfp =
-  Proofview.Goal.enter begin fun gl ->
-  let helpercsts = List.map (fun (cst, i) -> cst) info.helpers_info in
-  let opacify, transp = simpl_of ((destConstRef (Lazy.force coq_hidebody), Conv_oracle.transparent)
-    :: List.map (fun x -> x, Conv_oracle.Expand) (f_cst :: funf_cst :: helpercsts)) in
-  let opacified tac = wrap tac opacify transp in
-  let my_simpl = opacified simpl_in_concl in
   Proofview.tclORELSE (
     tclTHENLIST
-      [set_eos_tac (); intros; prove_unfolding_lemma_aux info where_map my_simpl [f_cst, funf_cst] p unfp] >>= fun () ->
+      [set_eos_tac (); intros; aux_program [f_cst, funf_cst] p unfp] >>= fun () ->
     let () = transp () in
     Proofview.tclUNIT ())
     (fun (e, info) -> let () = transp () in Proofview.tclZERO ~info e)
 
   end
 
-let prove_unfolding_sublemma info where_map f_cst funf_cst (subst, p, unfp) =
-  let helpercsts = List.map (fun (cst, i) -> cst) info.helpers_info in
-  let opacify, transp = simpl_of ((destConstRef (Lazy.force coq_hidebody), Conv_oracle.transparent)
-    :: List.map (fun x -> x, Conv_oracle.Expand) (f_cst :: funf_cst :: helpercsts)) in
-  let opacified tac = wrap tac opacify transp in
-  let my_simpl = opacified simpl_in_concl in
-  prove_unfolding_lemma_aux info where_map my_simpl subst p unfp
-
 let prove_unfolding_lemma info where_map f_cst funf_cst p unfp =
   Proofview.Goal.enter begin fun gl ->
   let () =
@@ -1253,7 +1148,7 @@ let ind_elim_tac indid inds mutinds info ind_fun =
                       Tactics.simpl_in_concl; Tactics.intros;
                       prove_methods (Reductionops.nf_beta env sigma app)]
        else
-         let ctx, concl = decompose_prod_decls sigma ty in
+         let ctx, concl = decompose_prod_assum sigma ty in
          Proofview.Unsafe.tclEVARS sigma <*>
          Tactics.simpl_in_concl <*> Tactics.intros <*>
          Tactics.cut concl <*>
@@ -1272,4 +1167,4 @@ let ind_elim_tac indid inds mutinds info ind_fun =
     | _, _ -> assert false)
   in
   try observe "applyind" (applyind inds [])
-  with e -> tclFAIL (Pp.str"exception")
+  with e -> tclFAIL 0 (Pp.str"exception")
