@@ -29,38 +29,37 @@ let mkConstructG c u =
 let mkIndG c u =
   mkIndU (destIndRef (Lazy.force c), u)
 
-let mkAppG env evd gr args =
+let mkAppG evd gr args = 
   let c = e_new_global evd gr in
-  evd_comb1 (fun sigma c -> Typing.checked_appvect env sigma c args) evd c
+    mkApp (c, args)
 
-let applistG env evd gr args =
-  mkAppG env evd gr (Array.of_list args)
+let applistG evd gr args = 
+  mkAppG evd gr (Array.of_list args)
 
-let mkSig env evd (n, c, t) =
+let mkSig evd (n, c, t) = 
   let args = [| c; mkLambda (n, c, t) |] in
-    mkAppG env evd (Lazy.force coq_sigma) args
+    mkAppG evd (Lazy.force coq_sigma) args
 
 let constrs_of_coq_sigma env evd t alias = 
   let rec aux env proj c ty =
     match kind !evd c with
-    | App (f, args) when is_global env !evd (Lazy.force coq_sigmaI) f && 
+    | App (f, args) when is_global !evd (Lazy.force coq_sigmaI) f && 
 	                   Array.length args = 4 ->
        let ty = Retyping.get_type_of env !evd args.(1) in
 	(match kind !evd ty with
-          | Prod (n, b, t) ->
-            (* sigma is not sort poly (at least for now) *)
-            let p1 = mkProj (Lazy.force coq_pr1, ERelevance.relevant, proj) in
-            let p2 = mkProj (Lazy.force coq_pr2, ERelevance.relevant, proj) in
+	| Prod (n, b, t) ->
+	    let p1 = mkProj (Lazy.force coq_pr1, proj) in
+	    let p2 = mkProj (Lazy.force coq_pr2, proj) in
 	    (n, args.(2), p1, args.(0)) ::
               aux (push_rel (of_tuple (n, None, b)) env) p2 args.(3) t
 	| _ -> raise (Invalid_argument "constrs_of_coq_sigma"))
-    | _ -> [(anonR, c, proj, ty)]
+    | _ -> [(Context.anonR, c, proj, ty)]
   in aux env alias t (Retyping.get_type_of env !evd t)
 
-let decompose_coq_sigma env sigma t = 
+let decompose_coq_sigma sigma t = 
   let s = Lazy.force coq_sigma in
     match kind sigma t with
-    | App (f, args) when is_global env sigma s f && Array.length args = 2 ->
+    | App (f, args) when is_global sigma s f && Array.length args = 2 ->
        let ind, u = destInd sigma f in
          Some (u, args.(0), args.(1))
     | _ -> None
@@ -114,11 +113,6 @@ let telescope_of_context env sigma ctx =
   let sigma, _ = Typing.type_of env sigma tele_interp in
   sigma, tele, tele_interp
 
-(* Given a context ⊢ Γ, returns
-  - ⊢ Tel(Γ) : Type (iterated Σ-type built out of Γ)
-  - p : Tel(Γ) ⊢ Ctx(Γ) (a context of let-bindings made of each projection in order)
-  - Γ ⊢ Intro(Γ) : Tel(Γ) (iterated constructors)
-*)
 let telescope env evd = function
   | [] -> assert false
   | [d] -> let (n, _, t) = to_tuple d in t, [of_tuple (n, Some (mkRel 1), Vars.lift 1 t)],
@@ -127,7 +121,7 @@ let telescope env evd = function
       let (n, _, t) = to_tuple d in
       let len = succ (List.length tl) in
       let ts = Retyping.get_sort_of (push_rel_context tl env) !evd t in
-      let ts = ESorts.kind !evd ts in
+      let tuniv = Sorts.univ_of_sort ts in
       let ty, tys =
         let rec aux (ty, tyuniv, tys) ds =
           match ds with
@@ -135,22 +129,21 @@ let telescope env evd = function
           | d :: ds ->
             let (n, b, t) = to_tuple d in
             let pred = mkLambda (n, t, ty) in
-            let env = push_rel_context ds env in
-            let sigty = mkAppG env evd (Lazy.force coq_sigma) [|t; pred|] in
+            let sigty = mkAppG evd (Lazy.force coq_sigma) [|t; pred|] in
             let _, u = destInd !evd (fst (destApp !evd sigty)) in
-            let _, ua = UVars.Instance.to_array (EInstance.kind !evd u) in
-            let l = Sorts.sort_of_univ @@ Univ.Universe.make ua.(0) in
+            let ua = Univ.Instance.to_array (EInstance.kind !evd u) in
+            let l = Univ.Universe.make ua.(0) in
+            let env = push_rel_context ds env in
             (* Ensure that the universe of the sigma is only >= those of t and pred *)
-            let open UnivProblem in
             let enforce_leq env sigma t cstr =
               let ts = Retyping.get_sort_of env sigma t in
-              let ts = ESorts.kind sigma ts in
-              UnivProblem.Set.add (ULe (ts, l)) cstr
+              let su = Sorts.univ_of_sort ts in
+              Univ.enforce_leq su l cstr
             in
-            let cstrs = enforce_leq env !evd t (UnivProblem.Set.add (ULe (tyuniv, l)) UnivProblem.Set.empty) in
-            let () = evd := Evd.add_universe_constraints !evd cstrs in
+            let cstrs = enforce_leq env !evd t (Univ.enforce_leq tyuniv l Univ.Constraints.empty) in
+            let () = evd := Evd.add_constraints !evd cstrs in
             aux (sigty, l, (u, pred) :: tys) ds
-        in aux (t, ts, []) tl
+        in aux (t, tuniv, []) tl
       in
       let constr, _ = 
 	List.fold_right (fun (u, pred) (intro, k) ->
@@ -165,9 +158,8 @@ let telescope env evd = function
       let (last, _, subst) = List.fold_right2
 	(fun pred d (prev, k, subst) ->
           let (n, b, t) = to_tuple d in
-          (* sigma is not sort poly (at least for now) *)
-          let proj1 = mkProj (Lazy.force coq_pr1, ERelevance.relevant, prev) in
-          let proj2 = mkProj (Lazy.force coq_pr2, ERelevance.relevant, prev) in
+	  let proj1 = mkProj (Lazy.force coq_pr1, prev) in
+	  let proj2 = mkProj (Lazy.force coq_pr2, prev) in
 	    (Vars.lift 1 proj2, succ k, of_tuple (n, Some proj1, Vars.liftn 1 k t) :: subst))
 	(List.rev tys) tl (mkRel 1, 1, [])
       in ty, (of_tuple (n, Some last, Vars.liftn 1 len t) :: subst), constr
@@ -175,7 +167,7 @@ let telescope env evd = function
 let sigmaize ?(liftty=0) env0 evd pars f =
   let env = push_rel_context pars env0 in
   let ty = Retyping.get_type_of env !evd f in
-  let ctx, concl = whd_decompose_prod_decls env !evd ty in
+  let ctx, concl = splay_prod_assum env !evd ty in
   let ctx = EConstr.Vars.smash_rel_context ctx in
   let argtyp, letbinders, make = telescope env evd ctx in
     (* Everyting is in env, move to index :: letbinders :: env *) 
@@ -188,7 +180,7 @@ let sigmaize ?(liftty=0) env0 evd pars f =
 	        letbinders)
   in
   let tyargs = [| argtyp; pred |] in
-  let tysig = mkAppG env evd (Lazy.force coq_sigma) tyargs in
+  let tysig = mkAppG evd (Lazy.force coq_sigma) tyargs in
   let indexproj = Lazy.force coq_pr1 in
   let valproj = Lazy.force coq_pr2 in
   let indices = 
@@ -204,8 +196,9 @@ let sigmaize ?(liftty=0) env0 evd pars f =
 			rel_vect 0 lenb))
 		 (Equations_common.lift_rel_contextn 1 (succ lenb) letbinders))
     in
-    let (_, u) = destInd !evd (fst @@ destApp !evd tysig) in
-    mkApp (mkRef (Lazy.force coq_sigmaI, u), [|argtyp; pred; Vars.lift 1 make; mkRel 1|])
+    let _tylift = Vars.lift lenb argtyp in
+      mkAppG evd (Lazy.force coq_sigmaI)
+	[|argtyp; pred; Vars.lift 1 make; mkRel 1|]
   in
   let pred = it_mkLambda_or_LetIn pred pars in
   let _ = e_type_of env0 evd pred in
@@ -222,7 +215,8 @@ let signature_class evd =
 
 let build_sig_of_ind env sigma (ind,u as indu) =
   let (mib, oib as _mind) = Inductive.lookup_mind_specif env ind in
-  let ctx = inductive_alldecls env indu in
+  let ctx = inductive_alldecls env (from_peuniverses sigma indu) in
+  let ctx = List.map of_rel_decl ctx in
   let ctx = EConstr.Vars.smash_rel_context ctx in
   let lenpars = mib.mind_nparams_rec in
   let lenargs = List.length ctx - lenpars in
@@ -347,7 +341,7 @@ let pattern_sigma ~assoc_right c hyp env sigma =
 	    constrs_of_coq_sigma env evd t p @ terms
          | _ -> terms
   in
-  let pat x = Patternops.pattern_of_constr env !evd x in
+  let pat x = Patternops.pattern_of_constr env !evd (to_constr !evd x) in
   let terms = 
     if assoc_right then terms
     else match terms with
@@ -375,7 +369,7 @@ let curry_left_hyp env sigma c t =
   let rec curry_index c t =
     match kind sigma t with
     | Prod (na, dom, concl) ->
-       (match decompose_coq_sigma env sigma dom with
+       (match decompose_coq_sigma sigma dom with
 	| None -> (c, t)
 	| Some (u, ty, pred) ->
 	   let term, typ = aux c t na u ty pred concl in
@@ -390,7 +384,7 @@ let curry_left_hyp env sigma c t =
   let curry c t =
     match kind sigma t with
     | Prod (na, dom, concl) ->
-       (match decompose_coq_sigma env sigma dom with
+       (match decompose_coq_sigma sigma dom with
 	| None -> None
 	| Some (inst, ty, pred) ->
 	   let term, typ = aux c t na inst ty pred concl in
@@ -399,11 +393,11 @@ let curry_left_hyp env sigma c t =
     | _ -> None
   in curry c t
 
-let curry env sigma na c =
+let curry sigma na c =
   let rec make_arg na t =
-    match decompose_coq_sigma env sigma t with
+    match decompose_coq_sigma sigma t with
     | None -> 
-       if is_global env sigma (Lazy.force logic_unit) t then
+       if is_global sigma (Lazy.force logic_unit) t then
          let _, u = destInd sigma t in
          [], constr_of_global_univ sigma (Lazy.force logic_unit_intro, u)
        else [of_tuple (na,None,t)], mkRel 1
@@ -430,13 +424,13 @@ let uncurry_hyps name =
     let sigma = Goal.sigma gl in
     let hyps, _ =
       List.split_when (fun d ->
-          is_global env sigma (Lazy.force coq_end_of_section) (get_named_type d)
+          is_global sigma (Lazy.force coq_end_of_section) (get_named_type d)
           || is_section_variable (Global.env ()) (get_id d)) hyps in
     let ondecl (sigma, acc, ty) d =
       let (dna, _, dty) = to_named_tuple d in
       let sigma, sigmaI = new_global sigma (Lazy.force coq_sigmaI) in
       let _, u = destConstruct sigma sigmaI in
-      let types = [| dty; mkNamedLambda sigma dna dty ty |] in
+      let types = [| dty; mkNamedLambda dna dty ty |] in
       let app = mkApp (sigmaI, Array.append types [| mkVar dna.binder_name; acc |]) in
       (sigma, app, mkApp (mkIndG coq_sigma u, types))
     in
@@ -453,10 +447,10 @@ let uncurry_hyps name =
 let uncurry_call env sigma fn c =
   let hd', args' = decompose_app sigma fn in
   let hd, args = decompose_app sigma c in
-  let params, args = Array.chop (Array.length args') args in
-  let hd = mkApp (hd, params) in
+  let params, args = List.chop (List.length args') args in
+  let hd = applist (hd, params) in
   let ty = Retyping.get_type_of env sigma hd in
-  let ctx, _ = Reductionops.whd_decompose_prod_decls env sigma ty in
+  let ctx, _ = Reductionops.splay_prod_assum env sigma ty in
   let ctx =
     let open Context.Rel.Declaration in
     let rec aux env ctx args =
@@ -470,14 +464,14 @@ let uncurry_call env sigma fn c =
         decl :: aux env decls args
       | [], _ :: _ -> assert false
       | ctx, [] -> []
-    in List.rev (aux env (List.rev ctx) (Array.to_list args))
+    in List.rev (aux env (List.rev ctx) args)
   in
   let evdref = ref sigma in
   if CList.is_empty ctx then 
     user_err_loc (None, Pp.str"No arguments to uncurry");
   (* let ctx = (Anonymous, None, concl) :: ctx in *)
   let sigty, sigctx, constr = telescope env evdref ctx in
-  let app = Vars.substl (Array.rev_to_list args) constr in
+  let app = Vars.substl (List.rev args) constr in
   let fnapp = mkApp (hd, rel_vect 0 (List.length sigctx)) in
   let fnapp = it_mkLambda_or_subst env fnapp sigctx in
   let projsid = nameR (Id.of_string "projs") in
@@ -514,16 +508,18 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
   let after, rel_decl, before = Context_map.split_context (pred rel) ctx in
   let rel_ty = Context.Rel.Declaration.get_type rel_decl in
   let rel_ty = Vars.lift rel rel_ty in
-  let rel_t = EConstr.mkRel rel in
+  let rel_t = Constr.mkRel rel in
   (* Fetch some information about the type of the variable being eliminated. *)
-  let pind, args = Inductiveops.find_inductive env !evd rel_ty in
-  let mib, oib = Global.lookup_inductive (fst pind) in
+  let pind, args = Inductive.find_inductive env (to_constr ~abort_on_undefined_evars:false !evd rel_ty) in
+  let mib, oib = Global.lookup_pinductive pind in
   let params, indices = List.chop mib.mind_nparams args in
   (* The variable itself will be treated for all purpose as one of its indices. *)
   let indices = indices @ [rel_t] in
   let indfam = Inductiveops.make_ind_family (pind, params) in
   let arity_ctx = Inductiveops.make_arity_signature env !evd true indfam in
   let rev_arity_ctx = List.rev arity_ctx in
+  let params = List.map of_constr params in
+  let indices = List.map of_constr indices in
 
   (* Firstly, we need to analyze each index to decide if we should introduce
    * an equality for it or not. *)
@@ -616,10 +612,9 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
   let subst = Context_map.id_subst (arity_ctx @ ctx) in
   let rev_subst = Context_map.id_subst (arity_ctx @ ctx) in
   let subst, rev_subst = List.fold_left (
-    fun (subst, rev_subst) -> function
+    fun ((ctx, _, _) as subst, rev_subst) -> function
     | None -> subst, rev_subst
     | Some rel ->
-        let ctx = subst.Context_map.src_ctx in
         let fresh_rel = Context_map.mapping_constr !evd subst (EConstr.mkRel 1) in
         let target_rel = EConstr.mkRel (rel + oib.mind_nrealargs + 1) in
         let target_rel = Context_map.mapping_constr !evd subst target_rel in
@@ -633,8 +628,8 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
    (Context_map.mapping_constr !evd subst (EConstr.mkRel 1))) in
   (* [ctx'] is the context under which we will build the case in a first step. *)
   (* This is [ctx] where everything omitted and cut is removed. *)
-  let ctx' = List.skipn (nb_cuts_omit + oib.mind_nrealargs + 1) (rev_subst.Context_map.tgt_ctx) in
-  let rev_subst' = List.skipn (nb_cuts_omit + oib.mind_nrealargs + 1) (rev_subst.Context_map.map_inst) in
+  let ctx' = List.skipn (nb_cuts_omit + oib.mind_nrealargs + 1) (pi3 rev_subst) in
+  let rev_subst' = List.skipn (nb_cuts_omit + oib.mind_nrealargs + 1) (pi2 rev_subst) in
   let rev_subst' = Context_map.lift_pats (-(oib.mind_nrealargs+1)) rev_subst' in
   let rev_subst_without_cuts = Context_map.mk_ctx_map env !evd ctx rev_subst' ctx' in
   (* Now we will work under a context with [ctx'] as a prefix, so we will be
@@ -642,12 +637,10 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
 
   (* ===== SUBSTITUTION ===== *)
   let subst = CList.fold_right_i (
-    fun i omit subst ->
+    fun i omit ((ctx, pats, _) as subst)->
     match omit with
     | None -> subst
     | Some rel ->
-        let ctx = subst.Context_map.src_ctx in
-        let pats = subst.Context_map.map_inst in
         let orig = oib.mind_nrealargs + 1 - i in
         let fresh_rel = Context_map.specialize !evd pats (Context_map.PRel orig) in
         let target_rel = EConstr.mkRel (rel + oib.mind_nrealargs + 1) in
@@ -666,7 +659,7 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
   (* Also useful: a substitution from [ctx] to the context with cuts. *)
   let subst_to_cuts =
     let lift_subst = Context_map.mk_ctx_map env !evd (arity_ctx @ ctx)
-    (Context_map.lift_pats (oib.mind_nrealargs + 1) (Context_map.id_pats ctx))
+    (Context_map.lift_pats (oib.mind_nrealargs + 1) (pi2 (Context_map.id_subst ctx)))
     ctx in
       Context_map.compose_subst ~unsafe:true env ~sigma:!evd subst lift_subst
   in
@@ -675,7 +668,7 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
   let goal = Context_map.mapping_constr !evd subst_to_cuts goal in
 
   (* ===== CUTS ===== *)
-  let cuts_ctx, remaining = List.chop nb_cuts (subst.Context_map.src_ctx) in
+  let cuts_ctx, remaining = List.chop nb_cuts (pi1 subst) in
   let fresh_ctx = List.firstn (oib.mind_nrealargs + 1) remaining in
   let revert_cut x =
     let rec revert_cut i = function
@@ -685,7 +678,7 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
          | Context.Rel.Declaration.LocalAssum _ -> Some (EConstr.mkRel i)
          | Context.Rel.Declaration.LocalDef _ -> None)
       | _ :: l -> revert_cut (succ i) l
-    in revert_cut (- oib.mind_nrealargs) (subst.Context_map.map_inst)
+    in revert_cut (- oib.mind_nrealargs) (pi2 subst)
   in
   let rev_cut_vars = CList.map revert_cut (CList.init nb_cuts (fun i -> succ i)) in
   let cut_vars = List.rev rev_cut_vars in
@@ -695,7 +688,7 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
   let goal, to_apply, simpl =
     if Int.equal nb 0 then goal, [], false
     else
-      let arity_ctx' = Context_map.specialize_rel_context !evd (subst_to_cuts.Context_map.map_inst) arity_ctx in
+      let arity_ctx' = Context_map.specialize_rel_context !evd (pi2 subst_to_cuts) arity_ctx in
       let rev_indices' = List.map (Context_map.mapping_constr !evd subst_to_cuts) rev_indices in
       let _, rev_sigctx, tele_lhs, tele_rhs =
         CList.fold_left3 (
@@ -711,7 +704,7 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
         ) (succ nb_cuts, [], [], []) arity_ctx' rev_indices' omitted
       in
       let sigctx = List.rev rev_sigctx in
-      let sigty, _, sigconstr = telescope (push_rel_context (subst.Context_map.src_ctx) env) evd sigctx in
+      let sigty, _, sigconstr = telescope (push_rel_context (pi1 subst) env) evd sigctx in
 
       (* Build a goal with an equality of telescopes at the front. *)
       let left_sig = Vars.substl (List.rev tele_lhs) sigconstr in
@@ -737,43 +730,46 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
   (* If something is wrong here, it means that one of the parameters was
    * omitted or cut, which should be wrong... *)
   let params = List.map (Vars.lift (-(nb_cuts + oib.mind_nrealargs + 1))) params in
-  let goal = EConstr.it_mkProd_or_LetIn goal cuts_ctx in
+  let goal = Termops.it_mkProd_or_LetIn goal cuts_ctx in
   let goal = it_mkLambda_or_LetIn goal fresh_ctx in
   let params = List.map (to_constr ~abort_on_undefined_evars:false !evd) params in
   let goal' = to_constr ~abort_on_undefined_evars:false !evd goal in
-  let branches_ty = Inductive.build_branches_type (from_peuniverses !evd pind) (mib, oib) params goal' in
+  let branches_ty = Inductive.build_branches_type pind (mib, oib) params goal' in
   (* Refresh the inductive family. *)
-  let indfam = Inductiveops.make_ind_family (pind, List.map EConstr.of_constr params) in
+  let indfam = Inductiveops.make_ind_family (pind, params) in
   let branches_info = Inductiveops.get_constructors env indfam in
   let full_subst =
-    let pats = Context_map.id_pats ctx in
+    let (ctx', pats, ctx) = Context_map.id_subst ctx in
     let pats = Context_map.lift_pats (oib.mind_nrealargs + 1) pats in
-    let ctx' = arity_ctx @ ctx in
+    let ctx' = arity_ctx @ ctx' in
       Context_map.mk_ctx_map env !evd ctx' pats ctx
   in
   let full_subst = Context_map.compose_subst ~unsafe:true env ~sigma:!evd subst full_subst in
-  let pats_ctx' = Context_map.id_pats ctx' in
-  let pats_cuts = Context_map.id_pats cuts_ctx in
+  let pats_ctx' = pi2 (Context_map.id_subst ctx') in
+  let pats_cuts = pi2 (Context_map.id_subst cuts_ctx) in
   let branches_subst = Array.map (fun summary ->
     (* This summary is under context [ctx']. *)
     let indices = summary.Inductiveops.cs_concl_realargs in
     let params = Array.of_list summary.Inductiveops.cs_params in
-    let term = EConstr.mkConstructU (summary.Inductiveops.cs_cstr) in
+    let params = Array.map of_constr params in
+    let indices = Array.map of_constr indices in
+    let term = EConstr.mkConstructU (to_peuniverses summary.Inductiveops.cs_cstr) in
     let term = EConstr.mkApp (term, params) in
     let term = Vars.lift (summary.Inductiveops.cs_nargs) term in
     let term = EConstr.mkApp (term, Context.Rel.instance EConstr.mkRel 0 summary.Inductiveops.cs_args) in
     (* Indices are typed under [args @ ctx'] *)
     let indices = (Array.to_list indices) @ [term] in
     let args = summary.Inductiveops.cs_args in
+    let args = List.map of_rel_decl args in
     (* Substitute the indices in [cuts_ctx]. *)
     let rev_indices = List.rev indices in
-    let pats_indices = List.map (Context_map.pat_of_constr env !evd) rev_indices in
+    let pats_indices = List.map (Context_map.pat_of_constr !evd) rev_indices in
     let pats_ctx' = Context_map.lift_pats summary.Inductiveops.cs_nargs pats_ctx' in
     let pats = pats_indices @ pats_ctx' in
     let cuts_ctx = Context_map.specialize_rel_context !evd pats cuts_ctx in
     let pats = Context_map.lift_pats nb_cuts pats in
     let pats = pats_cuts @ pats in
-    let csubst = Context_map.mk_ctx_map env !evd (cuts_ctx @ args @ ctx') pats (subst.Context_map.src_ctx) in
+    let csubst = Context_map.mk_ctx_map env !evd (cuts_ctx @ args @ ctx') pats (pi1 subst) in
       Context_map.compose_subst ~unsafe:true env ~sigma:!evd csubst full_subst
   ) branches_info in
   let branches_nb = Array.map (fun summary ->
@@ -799,25 +795,25 @@ let curry_hyp env sigma hyp t =
   let curry t =
     match kind sigma t with
     | Prod (na, dom, concl) ->
-       let ctx, arg = curry env sigma na dom in
+       let ctx, arg = curry sigma na dom in
        let term = mkApp (mkVar hyp, [| arg |]) in
        let ty = Reductionops.nf_betaiota env sigma (Vars.subst1 arg concl) in
        Some (it_mkLambda_or_LetIn term ctx, it_mkProd_or_LetIn ty ctx)
     | _ -> None
   in curry t
 
-open RedFlags
+open CClosure.RedFlags
 
 let red_curry () =
   let redpr pr =
-    fPROJ (Projection.repr (Lazy.force pr)) in
+    fCONST (Projection.constant (Lazy.force pr)) in
   let fold accu r = red_add accu r in
   let reds = mkflags [fDELTA; fBETA; fMATCH] in
   let reds = List.fold_left fold reds [redpr coq_pr1; redpr coq_pr2] in
   Reductionops.clos_norm_flags reds
 
 let curry_concl env sigma na dom codom =
-  let ctx, arg = curry env sigma na dom in
+  let ctx, arg = curry sigma na dom in
   let newconcl =
     let body = it_mkLambda_or_LetIn (Vars.subst1 arg codom) ctx in
     let inst = extended_rel_vect 0 ctx in
@@ -825,9 +821,8 @@ let curry_concl env sigma na dom codom =
   let proj last decl (terms, acc) =
     if last then (acc :: terms, acc)
     else
-      (* sigma is not sort poly (at least for now) *)
-      let term = mkProj (Lazy.force coq_pr1, ERelevance.relevant, acc) in
-      let acc = mkProj (Lazy.force coq_pr2, ERelevance.relevant, acc) in
+      let term = mkProj (Lazy.force coq_pr1, acc) in
+      let acc = mkProj (Lazy.force coq_pr2, acc) in
       (term :: terms, acc)
   in
   let terms, acc =
@@ -850,18 +845,18 @@ module Tactics =struct
 
   let curry_hyp id =
   Proofview.Goal.enter begin fun gl ->
-    let sigma = Proofview.Goal.sigma gl in
     let decl = pf_get_hyp id gl in
     let (na, body, ty) = to_named_tuple decl in
       match curry_hyp (pf_env gl) (project gl) id ty with
       | Some (prf, typ) ->
          (match body with
           | Some b ->
-             let newprf = Vars.replace_vars sigma [(id,b)] prf in
+             let newprf = Vars.replace_vars [(id,b)] prf in
              tclTHEN (clear [id]) (Tactics.letin_tac None (Name id) newprf (Some typ) nowhere)
           | None ->
-             assert_replacing id prf typ)
-      | None -> tclFAIL (str"No currying to do in " ++ Id.print id)
+             (tclTHENFIRST (assert_before_replacing id typ)
+                           (Logic.refiner ~check:false EConstr.Unsafe.(to_constr prf))))
+      | None -> tclFAIL 0 (str"No currying to do in " ++ Id.print id)
   end
 
   let curry =
@@ -873,7 +868,7 @@ module Tactics =struct
       | Prod (na, dom, codom) ->
          Refine.refine ~typecheck:true
            (fun sigma -> curry_concl env sigma na dom codom)
-      | _ -> Tacticals.tclFAIL (str"Goal cannot be curried") end
+      | _ -> Tacticals.tclFAIL 0 (str"Goal cannot be curried") end
 
   let uncurry_call c c' id id' =
     enter begin fun gl ->
@@ -895,7 +890,7 @@ module Tactics =struct
         Proofview.Unsafe.tclEVARS sigma' <*>
         letin_tac None (Name id') (beta_applist sigma (sigpack, [mkVar id])) None nowhere
       with Not_found ->
-        Tacticals.tclFAIL (str"No Signature instance found")
+        Tacticals.tclFAIL 0 (str"No Signature instance found")
     end
 
   let pattern_sigma id =
