@@ -156,12 +156,12 @@ let where_id w = w.where_program.program_info.program_id
 
 let where_context wheres =
   List.map (fun ({where_program; where_type } as w) ->
-             make_def (nameR (where_id w)) (Some (where_term w)) where_type) wheres
+             make_def (Context.nameR (where_id w)) (Some (where_term w)) where_type) wheres
 
 let where_program_type w =
   program_type w.where_program
   
-let pplhs env sigma lhs = pr_pats env sigma (lhs.Context_map.map_inst)
+let pplhs env sigma lhs = pr_pats env sigma (pi2 lhs)
   
 let pr_splitting_rhs ?(verbose=false) env' env'' sigma lhs rhs ty =
   let verbose pp = if verbose then pp else mt () in
@@ -176,7 +176,7 @@ let pr_program_info env sigma p =
   let open Pp in
   Names.Id.print p.program_id ++ str " : " ++
   Printer.pr_econstr_env env sigma (Syntax.program_type p) ++ str " : " ++
-  Printer.pr_econstr_env env sigma (mkSort (ESorts.make p.program_sort)) ++
+  Printer.pr_econstr_env env sigma (mkType p.program_sort) ++
   str " ( " ++
   (match p.program_rec with
    | Some (Structural ann) ->
@@ -194,7 +194,7 @@ let pr_splitting env sigma ?(verbose=false) split =
   let verb pp = if verbose then pp else mt () in
   let rec aux = function
     | Compute (lhs, wheres, ty, c) ->
-      let env' = push_rel_context (lhs.Context_map.src_ctx) env in
+      let env' = push_rel_context (pi1 lhs) env in
       let ppwhere w =
         hov 2 (str"where " ++ Id.print (where_id w) ++ str " : " ++
                (try Printer.pr_econstr_env env'  sigma w.where_type ++
@@ -212,7 +212,7 @@ let pr_splitting env sigma ?(verbose=false) split =
       (pr_splitting_rhs ~verbose env' env'' sigma lhs c ty ++ Pp.fnl () ++ ppwheres ++
        verb (hov 2 (fnl () ++ str "(in context: " ++ spc () ++ pr_context_map env sigma lhs ++ str")" ++ fnl ())))
     | Split (lhs, var, ty, cs) ->
-      let env' = push_rel_context (lhs.Context_map.src_ctx) env in
+      let env' = push_rel_context (pi1 lhs) env in
       (pplhs env' sigma lhs ++ str " split: " ++ pr_rel_name env' var ++
        Pp.fnl () ++
        verb (hov 2 (str" : " ++
@@ -234,7 +234,7 @@ let pr_splitting env sigma ?(verbose=false) split =
         info.refined_args,
         info.refined_revctx, info.refined_newprob, info.refined_newty
       in
-      let env' = push_rel_context (lhs.Context_map.src_ctx) env in
+      let env' = push_rel_context (pi1 lhs) env in
       hov 0 (pplhs env' sigma lhs ++ str " with " ++ Id.print id ++ str" := " ++
              Printer.pr_econstr_env env' sigma (mapping_constr sigma revctx c) ++ Pp.fnl () ++
              verb (hov 2 (str " : " ++ Printer.pr_econstr_env env' sigma cty ++ str" " ++
@@ -244,8 +244,8 @@ let pr_splitting env sigma ?(verbose=false) split =
                       hov 2 (str "refine args: " ++ prlist_with_sep spc (Printer.pr_econstr_env env' sigma)
                         info.refined_args) ++
                       hov 2 (str "New problem: " ++ pr_context_map env sigma newprob) ++ 
-                      hov 2 (str "For type: " ++ Printer.pr_econstr_env (push_rel_context (newprob.Context_map.src_ctx) env) sigma newty) ++
-                      hov 2 (str"Eliminating:" ++ pr_rel_name (push_rel_context (newprob.Context_map.src_ctx) env) (snd arg) ++ spc ()) ++
+                      hov 2 (str "For type: " ++ Printer.pr_econstr_env (push_rel_context (pi1 newprob) env) sigma newty) ++
+                      hov 2 (str"Eliminating:" ++ pr_rel_name (push_rel_context (pi1 newprob) env) (snd arg) ++ spc ()) ++
                       hov 2 (str "Revctx is: " ++ pr_context_map env sigma revctx) ++
                       hov 2 (str "New problem to problem substitution is: " ++ pr_context_map env sigma info.refined_newprob_to_lhs ++ Pp.fnl ()))) ++
              hov 0 (aux s))
@@ -348,9 +348,13 @@ let compute_possible_guardness_evidences sigma n fixbody fixtype =
          doesn't seem to worth the effort (except for huge mutual
          fixpoints ?) *)
     let m = Termops.nb_prod sigma fixtype in
-    let ctx = fst (decompose_prod_n_decls sigma m fixtype) in
+    let ctx = fst (decompose_prod_n_assum sigma m fixtype) in
     List.map_i (fun i _ -> i) 0 ctx
 
+let nf_fix sigma (nas, cs, ts) =
+  let inj c = EConstr.to_constr ~abort_on_undefined_evars:false sigma c in
+  (nas, Array.map inj cs, Array.map inj ts)
+    
 let define_mutual_nested env evd get_prog progs =
   let mutual =
     List.filter (fun (p, prog) -> not (is_nested p)) progs
@@ -371,7 +375,7 @@ let define_mutual_nested env evd get_prog progs =
       in
       let after = (nested - 1) - before in
       let fixb = (Array.make 1 idx, 0) in
-      let fixna = Array.make 1 (make_annot (Name p.program_id) (Retyping.relevance_of_sort (ESorts.make p.program_sort))) in
+      let fixna = Array.make 1 (nameR p.program_id) in
       let fixty = Array.make 1 (Syntax.program_type p) in
       (* Apply to itself *)
       let beforeargs = Termops.rel_list (signlen + 1) before in
@@ -449,8 +453,12 @@ let define_mutual_nested env evd get_prog progs =
     let possible_indexes =
       Array.map3 (compute_possible_guardness_evidences !evd) structargs bodies tys
     in
-    Pretyping.esearch_fix_guard env !evd (Array.to_list possible_indexes)
-      (names, tys, bodies)
+    let names, tys, bodies = nf_fix !evd (names, tys, bodies) in
+    try
+      Pretyping.search_guard env (Array.to_list possible_indexes)
+          (names, tys, bodies)
+    with e -> 
+      user_err_loc ((fst (List.hd progs)).program_loc, CErrors.print e)
   in
   let declare_fix_fns i (p,prog) =
     let newidx = indexes.(i) in
@@ -463,27 +471,14 @@ let define_mutual_nested env evd get_prog progs =
   let nested = List.map2 declare_nested nested nestedbodies in
   fixes, nested
 
-let check_typed ~where ?name env evd c =
-  let sigma, _ =
-    try Typing.type_of env evd c
-    with Type_errors.TypeError (env, tyerr) ->
-      anomaly Pp.(str where ++ spc () ++
-        str "Equations build an ill-typed term: " ++ Printer.pr_econstr_env env evd c ++
-        Himsg.explain_pretype_error env evd
-          (Pretype_errors.TypingError (Pretype_errors.of_type_error tyerr)))
-    | Pretype_errors.PretypeError (env, evd, tyerr) ->
-        anomaly Pp.(str where ++ spc () ++
-        str "Equations build an ill-typed term: " ++ Printer.pr_econstr_env env evd c ++
-        Himsg.explain_pretype_error env evd tyerr)
-  in
-  let check = Evd.check_constraints evd (snd @@ Evd.universe_context_set sigma) in
-  if not check then anomaly Pp.(str where ++ spc () ++ str "Equations missing constraints in " ++
-    str (Option.default "(anonymous)" name))
+let helper_evar evm evar env typ src =
+  let sign, typ', instance, _ = push_rel_context_to_named_context ~hypnaming:KeepExistingNames env evm typ in
+  let evm' = evar_declare sign evar typ' ~src evm in
+    evm', mkEvar (evar, instance)
 
 let term_of_tree env0 isevar sort tree =
   let rec aux env evm sort = function
-    | Compute (subst, where, ty, RProgram rhs) ->
-      let ctx = subst.Context_map.src_ctx in
+    | Compute ((ctx, _, _), where, ty, RProgram rhs) ->
       let compile_where ({where_program; where_type} as w)
           (env, evm, ctx) =
         let evm, c', ty' = evm, where_term w, where_type in
@@ -493,9 +488,8 @@ let term_of_tree env0 isevar sort tree =
       let body = it_mkLambda_or_LetIn rhs ctx and typ = it_mkProd_or_subst env evm ty ctx in
       evm, body, typ
 
-    | Compute (lhs, where, ty, REmpty (split, sp)) ->
+    | Compute ((ctx, _, _) as lhs, where, ty, REmpty (split, sp)) ->
        assert (List.is_empty where);
-       let ctx = lhs.Context_map.src_ctx in
        let evm, bot = new_global evm (Lazy.force logic_bot) in
        let evm, prf, _ = aux env evm sort (Split (lhs, split, bot, sp)) in
        let evm, case = new_global evm (Lazy.force logic_bot_case) in
@@ -504,16 +498,14 @@ let term_of_tree env0 isevar sort tree =
        let ty = it_mkProd_or_subst env evm ty ctx in
        evm, term, ty
 
-    | Mapping (subst, s) ->
-      let ctx = subst.Context_map.src_ctx in
+    | Mapping ((ctx, p, ctx'), s) ->
       let evm, term, ty = aux env evm sort s in
-      let args = Array.rev_of_list (snd (constrs_of_pats ~inacc_and_hide:false env evm subst.Context_map.map_inst)) in
+      let args = Array.rev_of_list (snd (constrs_of_pats ~inacc_and_hide:false env evm p)) in
       let term = it_mkLambda_or_LetIn (whd_beta env evm (mkApp (term, args))) ctx in
       let ty = it_mkProd_or_subst env evm (prod_appvect evm ty args) ctx in
       evm, term, ty
 
-    | Refined (subst, info, rest) ->
-      let ctx = subst.Context_map.src_ctx in
+    | Refined ((ctx, _, _), info, rest) ->
       let (id, _, _), ty, rarg, path, f, args, newprob, newty =
         info.refined_obj, info.refined_rettyp,
         info.refined_arg, info.refined_path,
@@ -526,8 +518,7 @@ let term_of_tree env0 isevar sort tree =
       let ty = it_mkProd_or_subst env evm ty ctx in
       evm, term, ty
 
-    | Split (subst, rel, ty, sp) ->
-      let ctx = subst.Context_map.src_ctx in
+    | Split ((ctx, _, _) as subst, rel, ty, sp) ->
       (* Produce parts of a case that will be relevant. *)
       let evm, block = Equations_common.(get_fresh evm coq_block) in
       let blockty = mkLetIn (anonR, block, Retyping.get_type_of env evm block, lift 1 ty) in
@@ -538,12 +529,12 @@ let term_of_tree env0 isevar sort tree =
 
       (* The next step is to use [simplify]. *)
       let simpl_step = if simpl then
-          Simplify.simplify [None, Simplify.Infer_many]
-        else Simplify.identity
+          Simplify.simplify [None, Simplify.Infer_many] env evd
+        else Simplify.identity env evd
       in
       let branches = Array.map2 (fun (ty, nb, csubst) next ->
         (* We get the context from the constructor arity. *)
-        let new_ctx, ty = EConstr.decompose_prod_n_decls !isevar nb ty in
+        let new_ctx, ty = EConstr.decompose_prod_n_assum !isevar nb ty in
         let new_ctx = Namegen.name_context env !isevar new_ctx in
         let envnew = push_rel_context (new_ctx @ ctx') env in
         (* Remove the cuts and append them to the context. *)
@@ -567,11 +558,7 @@ let term_of_tree env0 isevar sort tree =
           msg_debug(str"... named context:");
           msg_debug(Printer.pr_named_context env !evd (EConstr.Unsafe.to_named_context (named_context env)));
         end;
-        let _ =
-          let env = push_rel_context (cut_ctx @ new_ctx @ ctx') env in
-          evd_comb0 (fun sigma -> Typing.type_of env sigma ty) evd
-        in
-        let ((hole, c), _, lsubst) = Simplify.apply_simplification_fun simpl_step env evd (cut_ctx @ new_ctx @ ctx', ty, sort) in
+        let ((hole, c), lsubst) = simpl_step (cut_ctx @ new_ctx @ ctx', ty, sort) in
         if !debug then
           begin
             let open Feedback in
@@ -595,7 +582,7 @@ let term_of_tree env0 isevar sort tree =
             let next_subst = context_map_of_splitting s in
             let perm_subst = Context_map.make_permutation ~env evm subst next_subst in
             (* [next_term] starts with lambdas, so we apply it to its context. *)
-            let args = Equations_common.extended_rel_vect 0 (perm_subst.Context_map.tgt_ctx) in
+            let args = Equations_common.extended_rel_vect 0 (pi3 perm_subst) in
             let next_term = beta_appvect !evd next_term args in
             let next_term = Context_map.mapping_constr evm perm_subst next_term in
             (* We know the term is a correct instantiation of the evar, we
@@ -611,7 +598,7 @@ let term_of_tree env0 isevar sort tree =
             let vars_subst = List.map2 (fun decl c ->
                 let id = Context.Named.Declaration.get_id decl in
                 id, c) (Environ.named_context env) named in
-            let term = Vars.replace_vars !evd vars_subst next_term in
+            let term = Vars.replace_vars vars_subst next_term in
             let term = Vars.substl rels term in
             (* let _ =
              *   let env = Evd.evar_env ev_info in
@@ -632,7 +619,7 @@ let term_of_tree env0 isevar sort tree =
       (* Remove the block lets *)
       let rec clean_block c =
         match kind !evd c with
-        | LetIn (_, b, _, b') when Equations_common.is_global env !evd (Lazy.force coq_block) b ->
+        | LetIn (_, b, _, b') when Equations_common.is_global !evd (Lazy.force coq_block) b ->
           clean_block (subst1 b b')
         | _ -> EConstr.map !evd clean_block c
       in
@@ -647,15 +634,15 @@ let term_of_tree env0 isevar sort tree =
       let pind, args = find_inductive env !evd rel_ty in
 
       (* Build the case. *)
-      let case_info = Inductiveops.make_case_info env (fst pind) Constr.RegularStyle in
-      let indty = Inductiveops.find_rectype env !evd (mkApp (mkIndU pind, Array.of_list args)) in
+      let case_info = Inductiveops.make_case_info env (fst pind) elim_relevance Constr.RegularStyle in
+      let indty = Inductiveops.find_rectype env !evd (mkApp (mkIndU pind, Array.map_of_list EConstr.of_constr args)) in
       let case = Inductiveops.make_case_or_project env !evd indty case_info
-          (case_ty, elim_relevance) rel_t branches in
+          case_ty rel_t branches in
       let term = EConstr.mkApp (case, Array.of_list to_apply) in
       let term = EConstr.it_mkLambda_or_LetIn term ctx in
       let typ = it_mkProd_or_subst env evm ty ctx in
-      let () = if !Equations_common.debug then check_typed ~where:"splitting" env !evd term in
       let term = Evarutil.nf_evar !evd term in
+      evd := Typing.check env !evd term typ;
       !evd, term, typ
   in
   let evm, term, typ = aux env0 !isevar sort tree in
@@ -697,7 +684,7 @@ type program_shape =
 let make_program env evd p prob s rec_info =
   match rec_info with
   | Some r ->
-    let sort = ESorts.make p.program_sort in
+    let sort = p.program_sort in
     let lhs = id_subst r.rec_lets in
     (match r.rec_node with
      | WfRec wfr ->
@@ -707,8 +694,8 @@ let make_program env evd p prob s rec_info =
          | Some t -> t
        in
        let term = beta_appvect !evd wfr.wf_rec_term
-           [| beta_appvect !evd fn (extended_rel_vect 0 (lhs.Context_map.src_ctx)) |] in
-       Single (p, prob, rec_info, s, it_mkLambda_or_LetIn term (lhs.Context_map.src_ctx))
+           [| beta_appvect !evd fn (extended_rel_vect 0 (pi1 lhs)) |] in
+       Single (p, prob, rec_info, s, it_mkLambda_or_LetIn term (pi1 lhs))
      | StructRec sr ->
        let term, ty = term_of_tree env evd sort s in
        let args = extended_rel_vect 0 r.rec_lets in
@@ -730,7 +717,7 @@ let make_program env evd p prob s rec_info =
              | NestedOn None ->
                (match s with
                 | Split (ctx, var, _, _) ->
-                  NestedOn (Some ((List.length (ctx.Context_map.src_ctx)) - var - sr.struct_rec_protos, None))
+                  NestedOn (Some ((List.length (pi1 ctx)) - var - sr.struct_rec_protos, None))
                 | _ -> ann)
              | _ -> ann
            in
@@ -739,7 +726,7 @@ let make_program env evd p prob s rec_info =
        in
        Mutual (p', prob, r, s, after, (* lift 1  *)term))
   | None ->
-    Single (p, prob, rec_info, s, fst (term_of_tree env evd (ESorts.make p.program_sort) s))
+    Single (p, prob, rec_info, s, fst (term_of_tree env evd p.program_sort s))
 
 let update_rec_info p rec_info =
   match p.Syntax.program_rec, rec_info.rec_node with
@@ -827,16 +814,15 @@ let make_single_program env evd flags p prob s rec_info =
   | [p] -> p
   | _ -> raise (Invalid_argument "make_single_program: more than one program")
 
-let change_lhs s subs =
+let change_lhs s (l, p, r) =
   let open Context.Rel.Declaration in
   let l' =
     List.map
       (function LocalDef ({binder_name=Name id}, b, t) as decl ->
-         (try let b' = List.assoc id s in LocalDef (make_annot (Name id) (get_relevance decl), b', t)
+         (try let b' = List.assoc id s in LocalDef (nameR id, b', t)
           with Not_found -> decl)
-              | decl -> decl) subs.Context_map.src_ctx
-  in
-  { subs with Context_map.src_ctx = l' }
+              | decl -> decl) l
+  in l', p, r
 
 let change_splitting s sp =
   let rec aux = function
@@ -872,8 +858,8 @@ let check_splitting env evd sp =
       ()
     | Split (lhs, rel, ty, sp) ->
       let _ = check_ctx_map env evd lhs in
-      let _r = lookup_rel rel (push_rel_context (lhs.Context_map.src_ctx) env) in
-      let _ = check_type (lhs.Context_map.src_ctx) ty in
+      let _r = lookup_rel rel (push_rel_context (pi1 lhs) env) in
+      let _ = check_type (pi1 lhs) ty in
       Array.iter (Option.iter aux) sp
     | Mapping (lhs, sp) ->
       let _ = check_ctx_map env evd lhs in
@@ -890,7 +876,7 @@ let check_splitting env evd sp =
       let def = make_def (nameR (where_id w)) (Some (where_term w)) w.where_type in
       def :: ctx
     in
-    let ctx = List.fold_left check_where (lhs.Context_map.src_ctx) wheres in
+    let ctx = List.fold_left check_where (pi1 lhs) wheres in
     ctx
   and check_program p =
     let ty = program_type p in
@@ -920,7 +906,7 @@ let define_one_program_constants flags env0 isevar udecl unfold p =
       let evm, s' = aux env evm p.program_splitting in
       let isevar = ref evm in
       let env = Global.env () in
-      let term, ty = term_of_tree env isevar (ESorts.make p.program_info.program_sort) s' in
+      let term, ty = term_of_tree env isevar p.program_info.program_sort s' in
       let (cst, (evm, e)) =
         Equations_common.declare_constant (path_id (Id.of_string "functional" :: path))
           term (Some ty)
@@ -972,13 +958,12 @@ let define_one_program_constants flags env0 isevar udecl unfold p =
       let evm, s' = aux env evm s in
       evm, Mapping (lhs, s')
 
-    | Refined (lhs, info, rest) ->
-      let ctx = lhs.Context_map.src_ctx in
+    | Refined ((ctx, _, _) as lhs, info, rest) ->
       let evm', rest' = aux env evm rest in
       let isevar = ref evm' in
       let env = Global.env () in
-      let sort =
-          (Retyping.get_sort_of (push_rel_context ctx env) !isevar info.refined_rettyp) in
+      let sort = Sorts.univ_of_sort
+          (Retyping.get_sort_of (push_rel_context (pi1 lhs) env) !isevar info.refined_rettyp) in
       let t, ty = term_of_tree env isevar sort rest' in
       let (cst, (evm, e)) =
         Equations_common.declare_constant (path_id ~unfold info.refined_path)
@@ -1014,7 +999,7 @@ let is_comp_obl sigma comp hole_kind =
   | Some r ->
       match hole_kind, r with
       | ImplicitArg (GlobRef.ConstRef c, (n, _), _), (loc, id) ->
-        is_rec_call (snd r) c
+        is_rec_call sigma (snd r) (mkConst c)
       | _ -> false
 
 type term_info = {
@@ -1022,7 +1007,7 @@ type term_info = {
   term_ustate : UState.t;
   base_id : string;
   poly : bool;
-  scope : Locality.definition_scope;
+  scope : Locality.locality;
   decl_kind : Decls.definition_object_kind;
   helpers_info : (Constant.t * (int * int)) list;
   comp_obls : Constant.t list; (** The recursive call proof obligations *)
@@ -1039,12 +1024,10 @@ let warn_complete id =
   str "Equations definition " ++ Id.print id ++ str" is complete and requires no further proofs. " ++
   str "Use the \"Equations\" command to define it."
 
-let equations_cat = CWarnings.create_category ~name:"equations" ()
-
 let warn_complete = 
   CWarnings.create
     ~name:"equations-open-proof-complete"
-    ~category:equations_cat
+    ~category:"equations"
     ~default:CWarnings.Enabled
     warn_complete
 
@@ -1056,27 +1039,32 @@ let solve_equations_obligations ~pm flags recids loc i sigma hook =
   let types =
     List.map (fun (ev, evi) ->
         if !Equations_common.debug then
-          Feedback.msg_debug (str"evar type" ++ Printer.pr_econstr_env env sigma (Evd.evar_concl evi));
+          Feedback.msg_debug (str"evar type" ++ Printer.pr_econstr_env env sigma evi.Evd.evar_concl);
         let section_length = List.length (named_context env) in
         let evcontext = Evd.evar_context evi in
         let local_context, section_context =
           List.chop (List.length evcontext - section_length) evcontext
         in
-        let type_ = EConstr.it_mkNamedProd_or_LetIn sigma (Evd.evar_concl evi) local_context in
+        let type_ = Termops.it_mkNamedProd_or_LetIn evi.Evd.evar_concl local_context in
         let type_ = nf_beta env sigma type_ in
         env, ev, evi, local_context, type_) evars in
   (* Make goals from a copy of the evars *)
+  let isevar0 = ref sigma in
+  let wits = ref [] in
   let tele =
     let rec aux types evm =
       match types with
       | [] -> Proofview.TNil evm
       | (evar_env, ev, evi, local_context, type_) :: tys ->
-        let cont evm wit =
-          let evm = Evd.define ev (applist (wit, Context.Named.instance_list mkVar local_context)) evm in
-          aux tys evm
-        in
-        Proofview.TCons (evar_env, evm, nf_evar evm type_, cont)
-    in aux types sigma
+        let evar_env = nf_env_evar !isevar0 evar_env in
+        Proofview.TCons (evar_env, evm, nf_evar !isevar0 type_,
+           (fun evm' wit ->
+             isevar0 :=
+               Evd.define ev (applist (wit, Context.Named.instance_list mkVar local_context))
+                 !isevar0;
+             wits := wit :: !wits;
+             aux tys evm'))
+    in aux types (Evd.from_ctx (Evd.evar_universe_context sigma))
   in
   let do_intros =
     (* Force introductions to be able to shrink the bodies later on. *)
@@ -1109,24 +1097,17 @@ let solve_equations_obligations ~pm flags recids loc i sigma hook =
   pm, lemma
 
 let gather_fresh_context sigma u octx =
-  let (qvars, univs), _ = Evd.sort_context_set sigma in
-  let qarr, uarr = UVars.Instance.to_array u in
-  let qualities =
-    Array.fold_left (fun ctx' l ->
-        match l with
-        | Sorts.Quality.QConstant _ -> assert false
-        | QVar l ->
-          if not (Sorts.QVar.Set.mem l qvars) then Sorts.QVar.Set.add l ctx'
-          else ctx')
-      Sorts.QVar.Set.empty qarr
-  in
+  let ctx = Evd.universe_context_set sigma in
+  let univs = Univ.ContextSet.levels ctx in
+  let arr = Univ.Instance.to_array u in
   let levels =
     Array.fold_left (fun ctx' l ->
         if not (Univ.Level.Set.mem l univs) then Univ.Level.Set.add l ctx'
         else ctx')
-      Univ.Level.Set.empty uarr
+      Univ.Level.Set.empty arr
   in
-  (qualities, levels), (UVars.AbstractContext.instantiate u octx)
+  let ctx = Univ.ContextSet.of_set levels in
+  Univ.ContextSet.add_constraints (Univ.AbstractContext.instantiate u octx) ctx
 
 let swap (x, y) = (y, x)
 
@@ -1173,20 +1154,21 @@ let solve_equations_obligations_program ~pm flags recids loc i sigma hook =
       let hd, args = decompose_app !evd t in
       if EConstr.isRef !evd hd then
         (if !Equations_common.debug then
-           Feedback.msg_debug
-             (str"mapping obligation to " ++ Printer.pr_econstr_env env !evd t ++
-              Prettyp.print_about env !evd (CAst.make (Constrexpr.AN (Libnames.qualid_of_ident id))) None);
+           Feedback.msg_debug (str"mapping obligation to " ++ Printer.pr_econstr_env env !evd t ++
+                               Prettyp.print_name env !evd (CAst.make (Constrexpr.AN (Libnames.qualid_of_ident id))) None);
          let cst, i = EConstr.destConst !evd hd in
          let ctx = Environ.constant_context (Global.env ()) cst in
          let ctx = gather_fresh_context !evd (EConstr.EInstance.kind sigma i) ctx in
-         evd := Evd.merge_sort_context_set Evd.univ_flexible !evd ctx;
+         evd := Evd.merge_context_set Evd.univ_flexible !evd ctx;
          t)
       else t
     in
     let () =
       Evd.fold_undefined
       (fun ev evi () ->
-      let args = Evd.evar_identity_subst evi in
+      let args =
+        (List.map (fun d -> EConstr.mkVar (Context.Named.Declaration.get_id d))
+                       (Evd.evar_filtered_context evi)) in
       let evart = EConstr.mkEvar (ev, args) in
       let evc = cmap evc evart in
       evd := Evd.define ev (whd_beta env !evd (EConstr.of_constr evc)) !evd)
@@ -1196,7 +1178,7 @@ let solve_equations_obligations_program ~pm flags recids loc i sigma hook =
     let recobls =
       List.map (fun (id, c) ->
       (* Due to shrinking, we can get lambda abstractions first *)
-      let _, body = decompose_lambda_decls sigma (EConstr.of_constr c) in
+      let _, body = decompose_lam_assum sigma (EConstr.of_constr c) in
       let hd, _ = decompose_app sigma body in
       try fst (EConstr.destConst sigma hd)
       with Constr.DestKO -> assert false) obls
@@ -1205,14 +1187,14 @@ let solve_equations_obligations_program ~pm flags recids loc i sigma hook =
   in
   let obl_hook = Declare.Hook.make_g hook in
   let reduce x =
-    let flags = RedFlags.beta in
+    let flags = CClosure.beta in
     to_constr sigma (clos_norm_flags flags (Global.env ()) sigma (of_constr x))
   in
   let cinfo = Declare.CInfo.make ~name:oblsid ~typ:ty () in
   let info = Declare.Info.make ~poly ~scope ~kind () in
   let pm, _ =
     Declare.Obls.add_definition ~pm ~cinfo ~info 
-      ~obl_hook ~body:term ~uctx:(Evd.evar_universe_context sigma)
+      ~obl_hook ~term ~uctx:(Evd.evar_universe_context sigma)
       ~reduce ~opaque:false oblsinfo in
   pm
 
@@ -1229,7 +1211,7 @@ let simplify_evars evars t =
     | _ -> EConstr.map evars aux t
   in aux t
 
-let unfold_entry cst = Hints.HintsUnfoldEntry [Evaluable.EvalConstRef cst]
+let unfold_entry cst = Hints.HintsUnfoldEntry [Tacred.EvalConstRef cst]
 let add_hint local i cst =
   let locality = if local || Global.sections_are_opened () then Hints.Local else Hints.SuperGlobal in
   Hints.add_hints ~locality [Id.to_string i] (unfold_entry cst)
@@ -1241,7 +1223,7 @@ type 'a hook =
 let rec_type_ids =
   CList.map_append
     (function Some (Guarded l) -> List.map fst l
-            | Some (Logical (_, ids)) -> [snd ids]
+            | Some (Logical ids) -> [snd ids]
             | None -> [])
 
 let define_programs (type a) ~pm env evd udecl is_recursive fixprots flags ?(unfold=false) programs : a hook -> a * Declare.OblState.t * Declare.Proof.t option  =
@@ -1332,7 +1314,7 @@ let define_programs ~pm env evd udecl is_recursive fixprots flags ?(unfold=false
 let mapping_rhs sigma s = function
   | RProgram c -> RProgram (mapping_constr sigma s c)
   | REmpty (i, sp) ->
-      try match nth (s.Context_map.map_inst) (pred i) with
+      try match nth (pi2 s) (pred i) with
       | PRel i' -> REmpty (i', sp)
       | _ -> assert false
       with Not_found -> assert false
