@@ -90,15 +90,15 @@ and ('a, 'b) by_annot =
   | WellFounded of 'b
 
 and 'a where_clause = pre_prototype * 'a list
-and 'a wheres = 'a where_clause list * Vernacexpr.notation_declaration list
+and 'a wheres = 'a where_clause list * Vernacexpr.decl_notation list
 
 type program = (signature * clause list) list
 and signature = identifier * rel_context * constr (* f : Π Δ. τ *)
-and clause = Clause of Loc.t option * lhs * (clause, clause) rhs (* lhs rhs *)
+and clause = Loc.t option * lhs * (clause, clause) rhs (* lhs rhs *)
 
-type pre_equation = Pre_equation of Constrexpr.constr_expr input_pats * (pre_equation, pre_equation) rhs
+type pre_equation = Constrexpr.constr_expr input_pats * (pre_equation, pre_equation) rhs
 
-type pre_clause = Pre_clause of Loc.t option * lhs * (pre_equation, pre_clause) rhs
+type pre_clause = Loc.t option * lhs * (pre_equation, pre_clause) rhs
 
 type pre_equations = pre_equation where_clause list
 
@@ -158,7 +158,7 @@ and pr_proto env sigma ((_,id), _, _, l, t, ann) =
    | Some (WellFounded (t, rel)) -> str"by wf " ++ pr_constr_expr env sigma t ++ pr_opt (pr_constr_expr env sigma) rel
    | Some (Structural id) -> str"by struct " ++ pr_opt (fun x -> pr_id (snd x)) id)
 
-and pr_clause env sigma (Clause (loc, lhs, rhs)) =
+and pr_clause env sigma (loc, lhs, rhs) =
   pr_lhs env sigma lhs ++ pr_rhs env sigma rhs
 
 and pr_clauses env sigma =
@@ -192,7 +192,7 @@ and pr_prerhs_aux env sigma = function
 and pr_user_rhs env sigma = pr_opt (pr_user_rhs_aux env sigma)
 and pr_prerhs env sigma = pr_opt (pr_prerhs_aux env sigma)
 
-and pr_user_clause env sigma (Pre_equation (lhs, rhs)) =
+and pr_user_clause env sigma (lhs, rhs) =
   pr_user_lhs env sigma lhs ++ pr_user_rhs env sigma rhs
 
 and pr_user_clauses env sigma =
@@ -204,7 +204,7 @@ and pr_prewheres env sigma (l, nts) =
 and pr_prewhere env sigma (sign, eqns) =
   pr_proto env sigma sign ++ str "{" ++ pr_user_clauses env sigma eqns ++ str "}"
 
-and pr_preclause env sigma (Pre_clause (loc, lhs, rhs)) =
+and pr_preclause env sigma (loc, lhs, rhs) =
   pr_lhs env sigma lhs ++ pr_prerhs env sigma rhs
 
 and pr_preclauses env sigma =
@@ -265,7 +265,7 @@ let equations_tactic =
 
 type rec_type_item =
   | Guarded of (Id.t * rec_annot) list (* for mutual rec *)
-  | Logical of int * Id.t with_loc (* for nested wf rec: number of arguments before the side-condition *)
+  | Logical of Id.t with_loc (* for nested wf rec *)
 
 type rec_type = rec_type_item option list
 
@@ -274,10 +274,14 @@ let is_structural = function Some (Guarded _) :: _ -> true | _ -> false
 let has_logical rec_type =
   List.exists (function Some (Logical _) -> true | _ -> false) rec_type
 
-let is_rec_call id c =
-  let id' = Label.to_id (Constant.label c) in
-  Id.equal id id'
-  
+let is_rec_call sigma id f =
+  match EConstr.kind sigma f with
+  | Var id' -> Id.equal id id'
+  | Const (c, _) ->
+    let id' = Label.to_id (Constant.label c) in
+    Id.equal id id'
+  | _ -> false
+
 let free_vars_of_constr_expr fid c =
   let rec aux bdvars l = function
     | { CAst.v = CRef (qid, _) } when qualid_is_ident qid ->
@@ -290,7 +294,7 @@ let free_vars_of_constr_expr fid c =
            | Globnames.TrueGlobal gr ->
              if not (Globnames.isConstructRef gr) then Id.Set.add id l
              else l
-           | Globnames.Abbrev _ -> l
+           | Globnames.SynDef _ -> l
          with Not_found -> Id.Set.add id l)
     | { CAst.v = CNotation (_,(InConstrEntry, "?( _ )"), _) } -> l
     | c -> fold_constr_expr_with_binders (fun a l -> a::l) aux bdvars l c
@@ -310,7 +314,7 @@ type program_info = {
   program_loc : Loc.t option;
   program_id : Id.t;
   program_orig_type : EConstr.t; (* The original type *)
-  program_sort : Sorts.t; (* The sort of this type *)
+  program_sort : Univ.Universe.t; (* The sort of this type *)
   program_sign : EConstr.rel_context;
   program_arity : EConstr.t;
   program_rec : program_rec_info option;
@@ -319,9 +323,9 @@ type program_info = {
 }
 
 let map_universe f u =
-  let u' = f (EConstr.mkSort (EConstr.ESorts.make u)) in
+  let u' = f (EConstr.mkSort (Sorts.sort_of_univ u)) in
   match Constr.kind (EConstr.Unsafe.to_constr u') with
-  | Sort s -> s
+  | Sort s -> Sorts.univ_of_sort s
   | _ -> assert false
 
 let map_program_info f p =
@@ -336,7 +340,7 @@ let _chole c loc =
   let kn = Lib.make_kn c in
   let cst = Names.Constant.make kn kn in
   CAst.make ~loc
-  (CHole (Some (GImplicitArg (GlobRef.ConstRef cst, (0,None), false)))), None
+  (CHole (Some (ImplicitArg (GlobRef.ConstRef cst, (0,None), false)), Namegen.IntroAnonymous,None)), None
 
 let _check_linearity env sigma opats =
   let rec aux ids pats = 
@@ -355,7 +359,7 @@ let _check_linearity env sigma opats =
   in ignore (aux Id.Set.empty opats)
 
 let is_implicit_arg = function
-  | GImplicitArg _ -> true
+  | ImplicitArg _ -> true
   | _ -> false
 
 let pattern_of_glob_constr env sigma avoid patname gc =
@@ -380,7 +384,7 @@ let pattern_of_glob_constr env sigma avoid patname gc =
     PUCstr (c, nparams, List.map (DAst.map_with_loc (aux Anonymous)) l)
   and aux patname ?loc = function
     | GVar id -> PUVar (id, User)
-    | GHole k ->
+    | GHole (k,_,_) ->
       (match patname with
       | Name id when is_implicit_arg k -> PUVar (id, Implicit)
       | _ -> 
@@ -388,11 +392,11 @@ let pattern_of_glob_constr env sigma avoid patname gc =
         let n = next_ident_away id avoid in
         PUVar (n, Generated))
     | GRef (GlobRef.ConstructRef cstr,_) -> constructor ?loc cstr []
-    | GRef (GlobRef.ConstRef _ as c, _) when Environ.QGlobRef.equal env c (Lazy.force coq_bang) -> PUEmpty
+    | GRef (GlobRef.ConstRef _ as c, _) when GlobRef.equal c (Lazy.force coq_bang) -> PUEmpty
     | GApp (c, l) ->
       begin match DAst.get c with
         | GRef (GlobRef.ConstructRef cstr,_) -> constructor ?loc cstr l
-        | GRef (GlobRef.ConstRef _ as c, _) when Environ.QGlobRef.equal env c (Lazy.force coq_inacc) ->
+        | GRef (GlobRef.ConstRef _ as c, _) when GlobRef.equal c (Lazy.force coq_inacc) ->
           let inacc = List.hd (List.tl l) in
           PUInac inacc
         | _ ->
@@ -410,35 +414,30 @@ let pattern_of_glob_constr env sigma avoid patname gc =
 
 let program_type p = EConstr.it_mkProd_or_LetIn p.program_arity p.program_sign
 
-let interp_pat env sigma notations ~avoid p pat =
+let interp_pat env notations ~avoid p pat =
+  let sigma = Evd.from_env env in
   let vars = (Id.Set.elements avoid) (* (ids_of_pats [p])) *) in
   (* let () = Feedback.msg_debug (str"Variables " ++ prlist_with_sep spc pr_id vars) in *)
   let tys = List.map (fun _ -> EConstr.mkProp) vars in
-  let rlv = List.map (fun _ -> EConstr.ERelevance.relevant) vars in
   let impls = List.map (fun _ -> []) vars in
   (* let () = Feedback.msg_debug (str"Internalizing " ++ pr_constr_expr p) in *)
   let ienv = try compute_internalization_env env sigma Variable vars tys impls with Not_found ->
     anomaly (str"Building internalization environment")
   in
   let notations = List.map Metasyntax.prepare_where_notation notations in
-  let vars, rlv, tys, impls, ienv =
+  let vars, tys, impls, ienv =
     match p with
     | Some (p, _) ->
       let ty = program_type p in
-      let r = Retyping.relevance_of_type env sigma ty in
       let ienv =
         try compute_internalization_env env sigma ~impls:ienv Recursive [p.program_id] [ty] [p.program_impls]
         with Not_found -> anomaly (str"Building internalization environment")
       in
-      (p.program_id :: vars, r :: rlv, ty :: tys, p.program_impls :: impls, ienv)
-    | None -> (vars, rlv, tys, impls, ienv)
+      (p.program_id :: vars, ty :: tys, p.program_impls :: impls, ienv)
+    | None -> (vars, tys, impls, ienv)
   in
   let nctx =
-    List.map3 (fun id r ty ->
-        Context.Named.Declaration.LocalAssum
-          (Context.make_annot id (EConstr.Unsafe.to_relevance r),
-           EConstr.Unsafe.to_constr ty))
-      vars rlv tys
+    List.map2 (fun id ty -> Context.Named.Declaration.LocalAssum (Context.annotR id, EConstr.Unsafe.to_constr ty)) vars tys
   in
   let env = Environ.push_named_context nctx env in
   let gc =
@@ -501,7 +500,7 @@ let interleave_implicits impls pats =
     | [], pats -> pats
   in aux impls pats
 
-let interp_eqn env sigma notations p ~avoid eqn =
+let interp_eqn env notations p ~avoid eqn =
   let whereid = ref (Nameops.add_suffix p.program_id "_abs_where") in
   let patnames =
     List.rev_map (fun decl -> Context.Rel.Declaration.get_name decl) p.program_sign
@@ -511,8 +510,8 @@ let interp_eqn env sigma notations p ~avoid eqn =
         if Impargs.is_status_implicit a then Some (Impargs.name_of_implicit a) else None)
       p.program_implicits
   in
-  let interp_pat notations avoid = interp_pat env sigma notations ~avoid in
-  let rec aux notations avoid curpats (Pre_equation (pat, rhs)) =
+  let interp_pat notations avoid = interp_pat env notations ~avoid in
+  let rec aux notations avoid curpats (pat, rhs) =
     let loc, avoid, pats =
       match pat with
       | SignPats pat ->
@@ -534,9 +533,9 @@ let interp_eqn env sigma notations p ~avoid eqn =
           else curpats @ pats in
         loc, avoid, pats
     in
-    Pre_clause (loc, pats, Option.map (interp_rhs notations avoid pats) rhs)
-  and aux2 notations avoid (Pre_equation (pat, rhs)) =
-    Pre_equation (pat, Option.map (interp_rhs' notations avoid) rhs)
+    (loc, pats, Option.map (interp_rhs notations avoid pats) rhs)
+  and aux2 notations avoid (pat, rhs) =
+    (pat, Option.map (interp_rhs' notations avoid) rhs)
   and interp_rhs' notations avoid = function
     | Refine (c, eqs) ->
       let avoid = Id.Set.union avoid (ids_of_pats None c) in
@@ -604,7 +603,7 @@ let interp_eqn env sigma notations p ~avoid eqn =
        *       let c = CApp ((None, CAst.(make ~loc (CRef (qid', ie)))), args) in
        *       let arg = CAst.make ~loc (CApp ((None, CAst.make ~loc c), [chole id' loc])) in
        *       arg) *)
-      | CGenarg eqns when Genarg.has_type eqns (Genarg.rawwit wit_equations_list) ->
+      | CHole (k, i, Some eqns) when Genarg.has_type eqns (Genarg.rawwit wit_equations_list) ->
         let eqns = Genarg.out_gen (Genarg.rawwit wit_equations_list) eqns in
         let id = !whereid in
         let () = whereid := Nameops.increment_subscript id in
@@ -620,8 +619,8 @@ let interp_eqn env sigma notations p ~avoid eqn =
     !wheres, c'
   in aux notations avoid [] eqn
 
-let is_recursive i : pre_equation wheres -> bool = fun eqs ->
-  let rec occur_eqn (Pre_equation (_, rhs)) =
+let is_recursive i : 'a wheres -> bool = fun eqs ->
+  let rec occur_eqn (_, rhs) =
     match rhs with
     | Some (Program (c,w)) ->
       (match c with
@@ -633,7 +632,7 @@ let is_recursive i : pre_equation wheres -> bool = fun eqs ->
     | _ -> false
   and occur_eqns eqs = List.exists occur_eqn eqs
   and occurs_notations nts =
-    List.exists (fun nt -> occur_var_constr_expr i nt.Vernacexpr.ntn_decl_interp) nts
+    List.exists (fun nt -> occur_var_constr_expr i nt.Vernacexpr.decl_ntn_interp) nts
   and occurs eqs =
     List.exists (fun (_,eqs) -> occur_eqns eqs) (fst eqs) || occurs_notations (snd eqs)
   in occurs eqs
